# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Image building."""


import json
from pathlib import Path
from typing import Dict, Optional, Tuple

from aea.cli.utils.config import get_default_author_from_cli_config
from aea.configurations.data_types import Dependency, PublicId

from autonomy.constants import (
    AUTONOMY_IMAGE_NAME,
    AUTONOMY_IMAGE_VERSION,
    DEFAULT_DOCKER_IMAGE_AUTHOR,
    OAR_IMAGE,
)
from autonomy.data import DATA_DIR
from autonomy.deploy.constants import DOCKERFILES
from autonomy.deploy.generators.docker_compose.base import get_docker_client


def generate_dependency_flag_var(dependencies: Tuple[Dependency, ...]) -> str:
    """Generate dependency flag env var"""
    args = []
    for dep in dependencies:
        # Flag for `aea install` command
        args += ["-e"]
        args += dep.get_pip_install_args()
    return " ".join(args)


class ImageProfiles:  # pylint: disable=too-few-public-methods
    """Image build profiles."""

    CLUSTER = "cluster"
    DEVELOPMENT = "dev"
    PRODUCTION = "prod"
    ALL = (CLUSTER, DEVELOPMENT, PRODUCTION)


def build_image(  # pylint: disable=too-many-arguments,too-many-locals
    agent: PublicId,
    pull: bool = False,
    version: Optional[str] = None,
    image_author: Optional[str] = None,
    extra_dependencies: Optional[Tuple[Dependency, ...]] = None,
    dockerfile: Optional[Path] = None,
) -> None:
    """Command to build images from for skaffold deployment."""

    tag: str
    path: str
    buildargs: Dict[str, str]
    docker_client = get_docker_client()
    buildargs = {
        "AUTONOMY_IMAGE_NAME": AUTONOMY_IMAGE_NAME,
        "AUTONOMY_IMAGE_VERSION": AUTONOMY_IMAGE_VERSION,
        "AEA_AGENT": str(agent),
        "AUTHOR": get_default_author_from_cli_config(),
        "EXTRA_DEPENDENCIES": generate_dependency_flag_var(extra_dependencies or ()),
    }

    image_version = version or agent.hash
    path = str(DATA_DIR / DOCKERFILES / "agent")
    if dockerfile is not None:  # pragma: nocover
        path = str(Path(dockerfile).parent)

    tag = OAR_IMAGE.format(
        image_author=image_author or DEFAULT_DOCKER_IMAGE_AUTHOR,
        agent=agent.name,
        version=image_version,
    )

    stream = docker_client.api.build(
        path=path,
        tag=tag,
        buildargs=buildargs,
        pull=pull,
    )

    for stream_obj in stream:
        for line in stream_obj.decode().split("\n"):
            line = line.strip()
            if not line:
                continue
            stream_data = json.loads(line)
            if "stream" in stream_data:
                print("[docker]" + stream_data["stream"], end="")
            elif "errorDetail" in stream_data:
                raise ImageBuildFailed(
                    "Image build failed with error "
                    + stream_data["errorDetail"]["message"]
                )
            elif "aux" in stream_data:
                print("[docker]" + stream_data["aux"]["ID"], end="")
            elif "status" in stream_data:  # pragma: no cover
                print("[docker]" + stream_data["status"], end="")


class ImageBuildFailed(Exception):
    """Raise when there's an error while building the agent image."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Build images."""

from pathlib import Path
from typing import Optional, Tuple

import click
from aea.cli.utils.click_utils import (
    PublicIdParameter,
    PyPiDependency,
    reraise_as_click_exception,
)
from aea.configurations.data_types import Dependency, PublicId

from autonomy.cli.helpers.image import build_image as _build_image
from autonomy.cli.utils.click_utils import image_author_option
from autonomy.deploy.image import ImageBuildFailed


@click.command(name="build-image")
@click.argument(
    "agent",
    type=PublicIdParameter(),
    required=False,
)
@click.option(
    "--service-dir",
    type=click.Path(dir_okay=True),
    help="Path to service dir.",
)
@click.option(
    "-e",
    "--extra-dependency",
    "extra_dependencies",
    type=PyPiDependency(),
    help="Provide extra dependency.",
    multiple=True,
)
@click.option("--version", type=str, help="Specify tag version for the image.")
@click.option("--dev", is_flag=True, help="Build development image.", default=False)
@click.option("--pull", is_flag=True, help="Pull latest dependencies.", default=False)
@click.option(
    "-f",
    "--dockerfile",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        exists=False,
    ),
    help="Specify custom dockerfile for building the agent",
)
@image_author_option
def build_image(  # pylint: disable=too-many-arguments
    agent: Optional[PublicId],
    service_dir: Optional[Path],
    dockerfile: Optional[Path],
    extra_dependencies: Tuple[Dependency, ...],
    pull: bool = False,
    dev: bool = False,
    version: Optional[str] = None,
    image_author: Optional[str] = None,
) -> None:
    """Build runtime images for autonomous agents."""
    if dev:
        click.echo(
            "`--dev` flag is deprecated, development mode does not require you to build docker image manually anymore."
        )

    with reraise_as_click_exception(ImageBuildFailed, FileNotFoundError):
        _build_image(
            agent=agent,
            service_dir=service_dir,
            extra_dependencies=extra_dependencies,
            pull=pull,
            version=version,
            image_author=image_author,
            dockerfile=dockerfile,
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test service management."""

import binascii
from typing import List, Optional

import pytest
from aea.components.base import load_aea_package
from aea.configurations.data_types import PackageType
from aea.configurations.loader import load_configuration_object
from aea.contracts.base import Contract
from aea.crypto.base import Crypto
from aea.crypto.registries import make_crypto
from aea_test_autonomy.configurations import ETHEREUM_KEY_PATH_5
from aea_test_autonomy.fixture_helpers import registries_scope_class  # noqa: F401
from click.testing import Result
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3

from autonomy.chain.base import ServiceState, registry_contracts
from autonomy.chain.config import ChainType
from autonomy.chain.constants import CHAIN_PROFILES, ERC20_TOKEN_ADDRESS_LOCAL
from autonomy.chain.service import get_service_info
from autonomy.cli.helpers.chain import ServiceHelper

from tests.conftest import ROOT_DIR
from tests.test_autonomy.test_chain.base import (
    AGENT_ID,
    BaseChainInteractionTest,
    COST_OF_BOND_FOR_AGENT,
    DUMMY_SERVICE,
    NUMBER_OF_SLOTS_PER_AGENT,
    THRESHOLD,
    patch_subgraph,
)


DEFAULT_AGENT_INSTANCE_ADDRESS = (
    "0x976EA74026E726554dB657fA54763abd0C3a0aa9"  # a key from default hardhat keys
)
MAX_DEPLOY_ATTEMPTS = 5


class BaseServiceManagerTest(BaseChainInteractionTest):
    """Test service manager."""

    cli_options = ("service",)
    key_file = ETHEREUM_KEY_PATH_5

    def mint_service(
        self,
        number_of_slots_per_agent: int = NUMBER_OF_SLOTS_PER_AGENT,
        threshold: int = THRESHOLD,
    ) -> int:
        """Mint service component"""
        with patch_subgraph(
            response=[{"tokenId": AGENT_ID, "publicId": "dummy_author/dummy_agent"}]
        ):
            service_id = self.mint_component(
                package_id=DUMMY_SERVICE,
                service_mint_parameters=dict(
                    agent_ids=[AGENT_ID],
                    number_of_slots_per_agent=[number_of_slots_per_agent],
                    cost_of_bond_per_agent=[COST_OF_BOND_FOR_AGENT],
                    threshold=threshold,
                ),
            )

        assert isinstance(service_id, int)
        return service_id

    def activate_service(self, service_id: int) -> None:
        """Activate service"""
        self.service_manager.activate(service_id=service_id)

    def register_instances(
        self, service_id: int, agent_instance: Optional[str] = None
    ) -> None:
        """Register agent instance."""
        self.service_manager.register_instance(
            service_id=service_id,
            instances=[(agent_instance or make_crypto("ethereum").address)],
            agent_ids=[AGENT_ID],
        )

    def deploy_service(
        self,
        service_id: int,
        reuse_multisig: bool = False,
    ) -> None:
        """Deploy service."""
        self.service_manager.deploy(
            service_id=service_id,
            reuse_multisig=reuse_multisig,
        )

    def terminate_service(self, service_id: int) -> None:
        """Terminate service."""
        self.service_manager.terminate(service_id=service_id)

    def unbond_service(self, service_id: int) -> None:
        """Unbond service."""
        self.service_manager.unbond(service_id=service_id)


class TestServiceManager(BaseServiceManagerTest):
    """Test service manager."""

    def test_activate_service(self) -> None:
        """Test activate service."""

        def _run_command(service_id: str) -> Result:
            """Run command and return result"""
            return self.run_cli(
                commands=(
                    "activate",
                    service_id,
                    "--key",
                    str(self.key_file),
                )
            )

        result = _run_command("100")
        # This should fail since the service is not there yet
        assert result.exit_code == 1, result.output
        assert "Service does not exist" in result.stderr

        service_id = self.mint_service()

        result = _run_command(str(service_id))
        # first attempt should be succesfull
        assert result.exit_code == 0, result.output
        assert "Service activated succesfully" in result.output

        result = _run_command(str(service_id))
        # first attempt should fail since the service already active
        assert result.exit_code == 1, service_id
        assert "Service must be inactive" in result.stderr

    def test_register_instances(
        self,
    ) -> None:
        """Test register instance on a service"""

        agent_instance = make_crypto("ethereum")

        def _run_command(service_id: str) -> Result:
            """Run command and return result"""
            return self.run_cli(
                commands=(
                    "register",
                    service_id,
                    "--key",
                    str(self.key_file),
                    "-a",
                    "1",
                    "-i",
                    agent_instance.address,
                )
            )

        result = _run_command("100")
        assert result.exit_code == 1, result.output
        assert "Service does not exist" in result.stderr

        service_id = self.mint_service()
        result = _run_command(str(service_id))
        assert result.exit_code == 1, result.output
        assert "Service needs to be in active state" in result.stderr

        self.activate_service(
            service_id=service_id,
        )

        result = _run_command(str(service_id))
        assert result.exit_code == 0, result.stderr
        assert "Agent instance registered succesfully" in result.output

        result = _run_command(str(service_id))
        # Will fail becaue the agent instance is already registered
        assert result.exit_code == 1, result.stdout
        assert "AgentInstanceRegistered" in result.stderr

    def test_deploy(
        self,
    ) -> None:
        """Test register instance on a service"""

        def _run_command(service_id: str) -> Result:
            """Run command and return result"""
            return self.run_cli(
                commands=(
                    "deploy",
                    service_id,
                    "--key",
                    str(self.key_file),
                )
            )

        service_id = self.mint_service()
        self.activate_service(
            service_id=service_id,
        )

        result = _run_command(str(service_id))
        assert result.exit_code == 1, result.output
        assert "Service needs to be in finished registration state" in result.stderr

        for _ in range(NUMBER_OF_SLOTS_PER_AGENT):
            self.register_instances(
                service_id=service_id,
            )

        result = _run_command(str(service_id))

        assert result.exit_code == 0, result.stderr
        assert "Service deployed succesfully" in result.output

        result = self.run_cli(
            commands=(
                "deploy",
                str(service_id),
                "--key",
                str(self.key_file),
            )
        )

        # Will fail becaue the service is already deployed
        assert result.exit_code == 1, result.stdout
        assert "Service needs to be in finished registration state" in result.stderr

    def test_terminate(
        self,
    ) -> None:
        """Test terminate service."""

        def _run_command(service_id: str) -> Result:
            """Run command and return result"""
            return self.run_cli(
                commands=(
                    "terminate",
                    service_id,
                    "--key",
                    str(self.key_file),
                )
            )

        result = _run_command("100")
        assert result.exit_code == 1, result.output
        assert "Service does not exist" in result.stderr

        service_id = self.mint_service()

        result = _run_command(str(service_id))
        assert result.exit_code == 1, result.output
        assert "Service not active" in result.stderr

        self.activate_service(
            service_id=service_id,
        )
        for _ in range(NUMBER_OF_SLOTS_PER_AGENT):
            self.register_instances(
                service_id=service_id,
            )
        self.deploy_service(service_id=service_id)

        result = _run_command(str(service_id))

        assert result.exit_code == 0, result.stderr
        assert "Service terminated succesfully" in result.output

        result = _run_command(str(service_id))

        # Will fail becaue the service is already deployed
        assert result.exit_code == 1, result.stdout
        assert "Service already terminated" in result.stderr

    def test_unbond(
        self,
    ) -> None:
        """Test unbond service."""

        def _run_command(service_id: str) -> Result:
            """Run command and return result"""
            return self.run_cli(
                commands=(
                    "unbond",
                    service_id,
                    "--key",
                    str(self.key_file),
                )
            )

        result = _run_command("100")
        assert result.exit_code == 1, result.output
        assert "Service does not exist" in result.stderr

        service_id = self.mint_service()

        self.activate_service(
            service_id=service_id,
        )
        for _ in range(NUMBER_OF_SLOTS_PER_AGENT):
            self.register_instances(
                service_id=service_id,
            )
        self.deploy_service(service_id=service_id)

        result = _run_command(str(service_id))
        assert result.exit_code == 1, result.output
        assert "Service needs to be in terminated-bonded state" in result.stderr

        self.terminate_service(service_id=service_id)

        result = _run_command(str(service_id))

        assert result.exit_code == 0, result.stderr
        assert "Service unbonded succesfully" in result.output

        result = _run_command(str(service_id))

        # Will fail becaue the service is already deployed
        assert result.exit_code == 1, result.stdout
        assert "Service needs to be in terminated-bonded state" in result.stderr

    def test_info(
        self,
    ) -> None:
        """Test service info."""

        def _run_check(service_id: str, message: str) -> Result:
            """Run command and return result"""
            result = self.run_cli(commands=("info", service_id))
            assert result.exit_code == 0
            assert message in result.output

        _run_check("100", message=ServiceState.NON_EXISTENT.name)

        service_id = self.mint_service()
        _run_check(
            service_id=str(service_id), message=ServiceState.PRE_REGISTRATION.name
        )

        self.activate_service(service_id=service_id)
        _run_check(
            service_id=str(service_id), message=ServiceState.ACTIVE_REGISTRATION.name
        )

        for _ in range(NUMBER_OF_SLOTS_PER_AGENT):
            self.register_instances(
                service_id=service_id,
            )
        _run_check(
            service_id=str(service_id), message=ServiceState.FINISHED_REGISTRATION.name
        )

        self.deploy_service(service_id=service_id)
        _run_check(service_id=str(service_id), message=ServiceState.DEPLOYED.name)

        self.terminate_service(service_id=service_id)
        _run_check(
            service_id=str(service_id), message=ServiceState.TERMINATED_BONDED.name
        )

        self.unbond_service(service_id=service_id)
        _run_check(
            service_id=str(service_id), message=ServiceState.PRE_REGISTRATION.name
        )

    def test_dry_run(self) -> None:
        """Test dry run."""
        service_id = self.mint_service()
        result = self.run_cli(
            commands=(
                "--dry-run",
                "activate",
                str(service_id),
                "--key",
                str(self.key_file),
            )
        )
        assert result.exit_code == 0, result.stderr
        assert "=== Dry run output ===" in result.output


class TestERC20AsBond(BaseServiceManagerTest):
    """Test ERC20 token as bond."""

    def setup(self) -> None:
        """Setup test."""
        super().setup()
        erc20_instance = Contract.from_dir(
            ROOT_DIR / "packages" / "valory" / "contracts" / "erc20"
        ).get_instance(
            ledger_api=self.ledger_api,
            contract_address=ERC20_TOKEN_ADDRESS_LOCAL,
        )
        tx = erc20_instance.functions.mint(
            self.crypto.address, int(1e18)
        ).build_transaction(
            {
                "from": self.crypto.address,
                "gas": 200001,
                "gasPrice": 30000,
                "nonce": self.ledger_api.api.eth.get_transaction_count(
                    self.crypto.address
                ),
            }
        )
        stx = self.crypto.entity.sign_transaction(tx)
        tx_digext = self.ledger_api.api.eth.send_raw_transaction(stx.rawTransaction)
        while True:
            try:
                return self.ledger_api.api.eth.get_transaction_receipt(tx_digext)
            except Exception:  # nosec
                continue

    def mint(self) -> int:
        """Mint service with token"""
        with patch_subgraph(
            response=[{"tokenId": AGENT_ID, "publicId": "dummy_author/dummy_agent"}]
        ):
            service_id = self.mint_component(
                package_id=DUMMY_SERVICE,
                service_mint_parameters=dict(
                    agent_ids=[AGENT_ID],
                    number_of_slots_per_agent=[NUMBER_OF_SLOTS_PER_AGENT],
                    cost_of_bond_per_agent=[COST_OF_BOND_FOR_AGENT],
                    threshold=THRESHOLD,
                    token=ERC20_TOKEN_ADDRESS_LOCAL,
                ),
            )
        assert isinstance(service_id, int)
        return service_id

    def test_mint(self) -> None:
        """Test mint with ERC20 token as bond."""
        service_id = self.mint()
        assert (
            ServiceHelper(
                service_id=service_id,
                chain_type=ChainType.LOCAL,
                key=self.key_file,
            )
            .check_is_service_token_secured(token=ERC20_TOKEN_ADDRESS_LOCAL)
            .token_secured
            is True
        )

    def test_activate(self) -> None:
        """Test activate with token."""
        service_id = self.mint()
        result = self.run_cli(
            commands=(
                "activate",
                str(service_id),
                "--key",
                str(self.key_file),
                "--token",
                ERC20_TOKEN_ADDRESS_LOCAL,
            )
        )
        assert result.exit_code == 0, result.stderr

    def test_activate_failure(self) -> None:
        """Test activate with token."""
        service_id = self.mint()
        result = self.run_cli(
            commands=(
                "activate",
                str(service_id),
                "--key",
                str(self.key_file),
            )
        )

        assert result.exit_code == 1, result.stderr
        assert (
            "Service is token secured, please provice token address using `--token` flag"
            in result.stderr
        )

    def test_register_instances(self) -> None:
        """Test register instances with token."""
        service_id = self.mint()
        result = self.run_cli(
            commands=(
                "activate",
                str(service_id),
                "--key",
                str(self.key_file),
                "--token",
                ERC20_TOKEN_ADDRESS_LOCAL,
            )
        )
        assert result.exit_code == 0, result.stderr

        agent = make_crypto("ethereum")
        result = self.run_cli(
            commands=(
                "register",
                str(service_id),
                "--key",
                str(self.key_file),
                "-a",
                str(AGENT_ID),
                "-i",
                agent.address,
                "--token",
                ERC20_TOKEN_ADDRESS_LOCAL,
            )
        )
        assert result.exit_code == 0, result.stderr

    def test_register_instances_failure(self) -> None:
        """Test register instances with token."""
        service_id = self.mint()
        result = self.run_cli(
            commands=(
                "activate",
                str(service_id),
                "--key",
                str(self.key_file),
                "--token",
                ERC20_TOKEN_ADDRESS_LOCAL,
            )
        )
        assert result.exit_code == 0, result.stderr

        agent = make_crypto("ethereum")
        result = self.run_cli(
            commands=(
                "register",
                str(service_id),
                "--key",
                str(self.key_file),
                "-a",
                str(AGENT_ID),
                "-i",
                agent.address,
            )
        )
        assert result.exit_code == 1, result.output
        assert (
            "Service is token secured, please provice token address using `--token` flag"
            in result.stderr
        )


class TestServiceRedeploymentWithSameMultisig(BaseServiceManagerTest):
    """Test service deployment with same multisig."""

    def setup(self) -> None:
        """Setup class."""
        super().setup()
        self.ledger_api._api = Web3(
            HTTPProvider(
                endpoint_uri="http://127.0.0.1:8545",
                request_kwargs={
                    "timeout": 120,
                },
            ),
        )
        self.ledger_api.api.eth.default_account = self.crypto.address

    def fund(self, address: str, amount: int = 1) -> None:
        """Fund an address."""
        raw_tx = {
            "to": address,
            "from": self.crypto.address,
            "value": self.ledger_api.api.to_wei(amount, "ether"),
            "gas": 100000,
            "chainId": self.ledger_api.api.eth.chain_id,
            "gasPrice": self.ledger_api.api.eth.gas_price,
            "nonce": self.ledger_api.api.eth.get_transaction_count(self.crypto.address),
        }
        signed_tx = self.crypto.entity.sign_transaction(raw_tx)
        self.ledger_api.api.eth.send_raw_transaction(signed_tx.rawTransaction)

    def generate_and_fund_keys(self, n: int = 4) -> List[Crypto]:
        """Generate and fund keys."""
        keys = []
        for _ in range(n):
            crypto = make_crypto("ethereum")
            keys.append(crypto)
        self.fund(address=keys[0].address)
        return keys

    def remove_owners(self, multisig_address: str, owners: List[Crypto]) -> None:
        """Remove owners."""

        # Load packages in the memory.
        packages_dir = ROOT_DIR / "packages" / "valory"
        for package_type, package_path in (
            (
                PackageType.CONTRACT,
                packages_dir / "contracts" / "gnosis_safe_proxy_factory",
            ),
            (PackageType.CONTRACT, packages_dir / "contracts" / "gnosis_safe"),
            (PackageType.CONTRACT, packages_dir / "contracts" / "multisend"),
            (
                PackageType.SKILL,
                packages_dir / "skills" / "transaction_settlement_abci",
            ),
        ):
            config_obj = load_configuration_object(
                package_type=package_type, directory=package_path
            )
            config_obj.directory = package_path
            load_aea_package(configuration=config_obj)

        from packages.valory.contracts.gnosis_safe.contract import SafeOperation
        from packages.valory.contracts.multisend.contract import MultiSendOperation
        from packages.valory.skills.transaction_settlement_abci.payload_tools import (
            hash_payload_to_hex,
            skill_input_hex_to_payload,
        )

        multisend_address = CHAIN_PROFILES["local"]["multisend"]
        threshold = 1
        owner_to_swap = owners[0].address
        owners_to_remove = reversed(owners[1:])
        multisend_txs = []

        for owner in owners_to_remove:
            txd = registry_contracts.gnosis_safe.get_remove_owner_data(
                ledger_api=self.ledger_api,
                contract_address=multisig_address,
                owner=owner.address,
                threshold=threshold,
            ).get("data")
            multisend_txs.append(
                {
                    "operation": MultiSendOperation.CALL,
                    "to": multisig_address,
                    "value": 0,
                    "data": HexBytes(bytes.fromhex(txd[2:])),
                }
            )
        txd = registry_contracts.gnosis_safe.get_swap_owner_data(
            ledger_api=self.ledger_api,
            contract_address=multisig_address,
            old_owner=self.ledger_api.api.to_checksum_address(owner_to_swap),
            new_owner=self.ledger_api.api.to_checksum_address(self.crypto.address),
        ).get("data")
        multisend_txs.append(
            {
                "operation": MultiSendOperation.CALL,
                "to": multisig_address,
                "value": 0,
                "data": HexBytes(txd[2:]),
            }
        )

        multisend_txd = registry_contracts.multisend.get_tx_data(
            ledger_api=self.ledger_api,
            contract_address=multisend_address,
            multi_send_txs=multisend_txs,
        ).get("data")
        multisend_data = bytes.fromhex(multisend_txd[2:])

        safe_tx_hash = registry_contracts.gnosis_safe.get_raw_safe_transaction_hash(
            ledger_api=self.ledger_api,
            contract_address=multisig_address,
            to_address=multisend_address,
            value=0,
            data=multisend_data,
            safe_tx_gas=0,
            operation=SafeOperation.DELEGATE_CALL.value,
        ).get("tx_hash")[2:]

        payload_data = hash_payload_to_hex(
            safe_tx_hash=safe_tx_hash,
            ether_value=0,
            safe_tx_gas=0,
            to_address=multisend_address,
            data=multisend_data,
        )

        tx_params = skill_input_hex_to_payload(payload=payload_data)
        safe_tx_bytes = binascii.unhexlify(tx_params["safe_tx_hash"])
        owner_to_signature = {}
        for owner_crypto in owners:
            signature = owner_crypto.sign_message(
                message=safe_tx_bytes,
                is_deprecated_mode=True,
            )
            owner_to_signature[
                self.ledger_api.api.to_checksum_address(owner_crypto.address)
            ] = signature[2:]

        tx = registry_contracts.gnosis_safe.get_raw_safe_transaction(
            ledger_api=self.ledger_api,
            contract_address=multisig_address,
            sender_address=owners[0].address,
            owners=tuple([owner.address for owner in owners]),
            to_address=tx_params["to_address"],
            value=tx_params["ether_value"],
            data=tx_params["data"],
            safe_tx_gas=tx_params["safe_tx_gas"],
            signatures_by_owner=owner_to_signature,
            operation=SafeOperation.DELEGATE_CALL.value,
        )
        stx = owners[0].sign_transaction(tx)
        tx_digest = self.ledger_api.send_signed_transaction(stx)
        self.ledger_api.get_transaction_receipt(tx_digest)

    @pytest.mark.parametrize(
        argnames="n_owners",
        argvalues=(1, 2),
    )
    def test_redeploy(self, n_owners: int) -> None:
        """Test redeploy service with same multisig."""
        instances = self.generate_and_fund_keys(n=n_owners)
        service_id = self.mint_service(
            number_of_slots_per_agent=n_owners, threshold=n_owners
        )
        self.activate_service(service_id=service_id)
        for instance in instances:
            self.register_instances(
                service_id=service_id, agent_instance=instance.address
            )
        self.deploy_service(service_id=service_id)
        self.terminate_service(service_id=service_id)
        self.unbond_service(service_id=service_id)
        _, multisig_address, *_ = get_service_info(
            ledger_api=self.ledger_api,
            chain_type=ChainType.LOCAL,
            token_id=service_id,
        )
        self.remove_owners(multisig_address=multisig_address, owners=instances)

        new_instances = self.generate_and_fund_keys(n=n_owners)
        self.activate_service(service_id=service_id)
        for instance in new_instances:
            self.register_instances(
                service_id=service_id, agent_instance=instance.address
            )

        self.deploy_service(service_id=service_id, reuse_multisig=True)
        _, multisig_address_redeployed, *_ = get_service_info(
            ledger_api=self.ledger_api,
            chain_type=ChainType.LOCAL,
            token_id=service_id,
        )
        assert multisig_address_redeployed == multisig_address
        safe_owners = registry_contracts.gnosis_safe.get_owners(
            ledger_api=self.ledger_api,
            contract_address=multisig_address,
        ).get("owners")
        assert set(safe_owners) == set([instance.address for instance in new_instances])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'offend_abci' skill."""

import json
from copy import deepcopy
from typing import Dict, Generator, Set, Type, cast

from packages.valory.skills.abstract_round_abci.base import (
    NUMBER_OF_BLOCKS_TRACKED,
    NUMBER_OF_ROUNDS_TRACKED,
    OffenceStatus,
    OffenseStatusEncoder,
)
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.offend_abci.models import OffendParams, SharedState
from packages.valory.skills.offend_abci.payloads import OffencesPayload
from packages.valory.skills.offend_abci.rounds import OffendAbciApp, OffendRound


class OffendBehaviour(BaseBehaviour):
    """Simulate offences."""

    matching_round = OffendRound

    @property
    def shared_state(self) -> SharedState:
        """Get the round sequence from the shared state."""
        return cast(SharedState, self.context.state)

    @property
    def offence_status(self) -> Dict[str, OffenceStatus]:
        """Get the offence status from the round sequence."""
        return self.shared_state.round_sequence.offence_status

    @property
    def first_agent(self) -> str:
        """Short the agents' addresses and get the first address."""
        return sorted(self.shared_state.synchronized_data.all_participants)[0]

    @property
    def params(self) -> OffendParams:
        """Return the params."""
        return cast(OffendParams, self.context.params)

    def _updated_status(self) -> Dict[str, OffenceStatus]:
        """Return the offense status updated."""
        status_per_agent = deepcopy(self.offence_status)
        affected_agent = self.first_agent
        status = status_per_agent[affected_agent]

        if self.params.validator_downtime:
            for _ in range(NUMBER_OF_BLOCKS_TRACKED):
                status.validator_downtime.add(True)

        for _ in range(NUMBER_OF_ROUNDS_TRACKED):
            if self.params.invalid_payload:
                status.invalid_payload.add(True)
            if self.params.blacklisted:
                status.blacklisted.add(True)
            if self.params.suspected:
                status.suspected.add(True)

        status.num_unknown_offenses = self.params.num_unknown
        status.num_double_signed = self.params.num_double_signed
        status.num_light_client_attack = self.params.num_light_client_attack

        return status_per_agent

    def async_act(self) -> Generator:
        """Create offences."""
        updated_status = self._updated_status()
        serialized_status = json.dumps(
            updated_status, cls=OffenseStatusEncoder, sort_keys=True
        )
        payload = OffencesPayload(self.context.agent_address, serialized_status)
        yield from self.send_a2a_transaction(payload)
        yield from self.wait_until_round_end()
        self.set_done()


class OffendRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the Offend abci app."""

    initial_behaviour_cls = OffendBehaviour
    abci_app_cls = OffendAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        OffendBehaviour,  # type: ignore
    }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Script to generate a markdown package table."""

from pathlib import Path

import requests


DOC_FILE = "docs/advanced_reference/on_chain_addresses.md"
ADDRESS_FILE_URL = "https://raw.githubusercontent.com/valory-xyz/autonolas-registries/main/docs/configuration.json"
CONTRACT_TO_SLUG = {
    "ComponentRegistry": "COMPONENT_REGISTRY",
    "AgentRegistry": "AGENT_REGISTRY",
    "RegistriesManager": "REGISTRIES_MANAGER",
    "ServiceRegistry": "SERVICE_REGISTRY",
    "ServiceRegistryL2": "SERVICE_REGISTRY",
    "ServiceRegistryTokenUtility": "SERVICE_REGISTRY_TOKEN_UTILITY",
    "ServiceManagerToken": "SERVICE_MANAGER",
    "ServiceManager": "SERVICE_MANAGER",
    "GnosisSafeMultisig": "GNOSIS_SAFE_PROXY_FACTORY",
    "GnosisSafeSameAddressMultisig": "GNOSIS_SAFE_SAME_ADDRESS_MULTISIG",
}
BLOCKSCAN_URLS = {
    "polygon": "https://polygonscan.com/address/",
    "polygonMumbai": "https://mumbai.polygonscan.com/address/",
    "gnosis": "https://gnosisscan.io/address/",
}


def to_title(string: str) -> str:
    """Convert camelCase to snake_case."""
    _string = string[0].upper()
    for char in string[1:]:
        if char.islower():
            _string += char
        else:
            _string += " "
            _string += char
    return _string


def main() -> None:
    """Generate contract addresses list."""
    data = "# List of contract addresses\n"
    chains = requests.get(ADDRESS_FILE_URL).json()
    for chain in chains:
        chain_name = chain["name"]
        chain_id = chain["chainId"]
        if chain_id in ("1", "5"):
            continue
        blockscan_url = BLOCKSCAN_URLS[chain_name]
        contracts = chain["contracts"]
        table = "| Name | Environment Variable | Address |\n"
        table += "| ---- | -------------------- | ------- |\n"
        for contract in contracts:
            name = contract["name"]
            if name not in CONTRACT_TO_SLUG:
                continue
            spaced_title = to_title(name)
            env_var = CONTRACT_TO_SLUG[name]
            address = contract["address"]
            table += f"| `{spaced_title}` | `CUSTOM_{env_var}_ADDRESS` | [`{address}`]({blockscan_url}{address}) |\n"

        # Multisend address is similar for all chains
        table += f"| `Multisend` | `CUSTOM_MULTISEND_ADDRESS` | [`0x40A2aCCbd92BCA938b02010E17A5b8929b49130D`]({blockscan_url}0x40A2aCCbd92BCA938b02010E17A5b8929b49130D) |\n"
        section = f"\n## {to_title(chain_name)} ({chain_id})\n"
        section += table
        data += section
    Path(DOC_FILE).write_text(data, encoding="utf-8")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""e2e tests for the `valory/offend_slash` skill."""

from pathlib import Path

import pytest
from aea.configurations.data_types import PublicId
from aea_test_autonomy.base_test_classes.agents import (
    BaseTestEnd2End,
    BaseTestEnd2EndExecution,
    RoundChecks,
)
from aea_test_autonomy.fixture_helpers import (  # noqa: F401  pylint: disable=unused-import
    UseACNNode,
    UseRegistries,
    abci_host,
    abci_port,
    acn_config,
    acn_node,
    flask_tendermint,
    hardhat_addr,
    hardhat_port,
    ipfs_daemon,
    ipfs_domain,
    key_pairs,
    registries_scope_class,
    tendermint_port,
)

from packages.valory.skills.offend_abci.rounds import OffendRound
from packages.valory.skills.registration_abci.rounds import RegistrationStartupRound
from packages.valory.skills.reset_pause_abci.rounds import ResetAndPauseRound
from packages.valory.skills.slashing_abci.rounds import StatusResetRound
from packages.valory.skills.transaction_settlement_abci.rounds import (
    RandomnessTransactionSubmissionRound,
    ValidateTransactionRound,
)


NO_SLASHING_HAPPY_PATH = (
    RoundChecks(RegistrationStartupRound.auto_round_id()),
    RoundChecks(OffendRound.auto_round_id(), n_periods=2),
    RoundChecks(ResetAndPauseRound.auto_round_id(), n_periods=2),
)

SLASHING_HAPPY_PATH = NO_SLASHING_HAPPY_PATH + (
    RoundChecks(ValidateTransactionRound.auto_round_id()),
)

SLASHING_STRICT_CHECKS = (
    "The Event.SLASH_START event was produced, "
    f"transitioning to `{RandomnessTransactionSubmissionRound.auto_round_id()}`",
    f"Entered in the '{StatusResetRound.auto_round_id()}' round for period 0",
    "The Event.SLASH_END event was produced. Switching back to the normal FSM.",
)


@pytest.mark.parametrize("nb_nodes", (4,))
class SlashingE2E(UseRegistries, UseACNNode, BaseTestEnd2End):
    """Test that slashing works right."""

    package_registry_src_rel = Path(__file__).parents[4]
    agent_package = "valory/offend_slash:0.1.0"
    skill_package = "valory/offend_slash_abci:0.1.0"
    wait_to_finish = 120
    _args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"

    def __set_configs(  # pylint: disable=unused-private-member
        self, i: int, nb_agents: int
    ) -> None:
        """Set the current agent's config overrides."""
        super().__set_configs(i=i, nb_agents=nb_agents)

        self.set_config(
            dotted_path=f"{self._args_prefix}.tendermint_p2p_url",
            value=f"localhost:{self._tendermint_image.get_p2p_port(i=i)}",
            type_="str",
        )


@pytest.mark.e2e
class TestSlashingThresholdUnmet(SlashingE2E, BaseTestEnd2EndExecution):
    """Test that slashing works right."""

    happy_path = NO_SLASHING_HAPPY_PATH
    extra_configs = [
        {
            "dotted_path": f"{SlashingE2E._args_prefix}.validator_downtime",
            "value": True,
        }
    ]


@pytest.mark.e2e
class TestSlashing(SlashingE2E, BaseTestEnd2EndExecution):
    """Test that slashing works right."""

    happy_path = SLASHING_HAPPY_PATH
    strict_check_strings = SLASHING_STRICT_CHECKS
    extra_configs = [
        {
            "dotted_path": f"{SlashingE2E._args_prefix}.validator_downtime",
            "value": True,
        },
        {
            "dotted_path": f"{SlashingE2E._args_prefix}.num_double_signed",
            "value": 1,
        },
    ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tx settlement helper."""

import time
from datetime import datetime
from typing import Callable, Dict, Optional, cast

from aea.configurations.data_types import PublicId
from aea.crypto.base import Crypto, LedgerApi
from requests.exceptions import ConnectionError as RequestsConnectionError

from autonomy.chain.base import registry_contracts
from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.exceptions import (
    ChainInteractionError,
    ChainTimeoutError,
    RPCError,
    TxBuildError,
)


DEFAULT_ON_CHAIN_INTERACT_TIMEOUT = 60.0
DEFAULT_ON_CHAIN_INTERACT_RETRIES = 5.0
DEFAULT_ON_CHAIN_INTERACT_SLEEP = 3.0

ERRORS_TO_RETRY = (
    "FeeTooLow",
    "wrong transaction nonce",
    "INTERNAL_ERROR: nonce too low",
    "Got empty transaction",
    "AlreadyKnown",
    "ALREADY_EXISTS",
    "already known",
    "ReplacementNotAllowed",
    "OldNonce",
)


def should_rebuild(error: str) -> bool:
    """Check if we should rebuild the transaction."""
    for _error in ("wrong transaction nonce", "OldNonce"):
        if _error in error:
            return True
    return False


def should_retry(error: str) -> bool:
    """Check an error message to check if we should raise an error or retry the tx"""
    if "Transaction with hash" in error and "not found" in error:
        return True
    for _error in ERRORS_TO_RETRY:
        if _error in error:
            return True
    return False


def should_reprice(error: str) -> bool:
    """Check an error message to check if we should reprice the transaction"""
    return "FeeTooLow" in error or "ReplacementNotAllowed" in error


class TxSettler:
    """Tx settlement helper"""

    tx: Optional[Dict]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        ledger_api: LedgerApi,
        crypto: Crypto,
        chain_type: ChainType,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        sleep: Optional[float] = None,
    ) -> None:
        """Initialize object."""
        self.chain_type = chain_type
        self.ledger_api = ledger_api
        self.crypto = crypto
        self.tx = None
        self.timeout = timeout or DEFAULT_ON_CHAIN_INTERACT_TIMEOUT
        self.retries = retries or DEFAULT_ON_CHAIN_INTERACT_RETRIES
        self.sleep = sleep or DEFAULT_ON_CHAIN_INTERACT_SLEEP

    def build(
        self,
        method: Callable[[], Dict],
        contract: str,
        kwargs: Dict,
    ) -> Dict:
        """Build transaction."""
        return method(  # type: ignore
            ledger_api=self.ledger_api,
            contract_address=ContractConfigs.get(name=contract).contracts[
                self.chain_type
            ],
            raise_on_try=True,
            **kwargs,
        )

    def _reprice(self, tx_dict: Dict) -> Optional[Dict]:
        """Reprice transaction."""
        if "maxFeePerGas" not in tx_dict or "maxPriorityFeePerGas" not in tx_dict:
            # This means something went wrong when building the transaction
            # returning a None value to the main loop will tell the main loop
            # to rebuild the transaction
            return None

        old_price = {
            "maxFeePerGas": tx_dict[  # pylint: disable=unsubscriptable-object
                "maxFeePerGas"
            ],
            "maxPriorityFeePerGas": tx_dict[  # pylint: disable=unsubscriptable-object
                "maxPriorityFeePerGas"
            ],
        }
        tx_dict.update(
            self.ledger_api.try_get_gas_pricing(
                old_price=old_price,
            )
        )
        return tx_dict

    @staticmethod
    def _already_known(error: str) -> bool:
        """Check if the transaction is alreade sent"""
        return "AlreadyKnown" in error

    def transact(
        self,
        method: Callable[[], Dict],
        contract: str,
        kwargs: Dict,
        dry_run: bool = False,
    ) -> Dict:
        """Make a transaction and return a receipt"""
        retries = 0
        tx_dict = None
        tx_digest = None
        already_known = False
        deadline = datetime.now().timestamp() + self.timeout
        while retries < self.retries and deadline >= datetime.now().timestamp():
            retries += 1
            try:
                if not already_known:
                    tx_dict = tx_dict or self.build(
                        method=method,
                        contract=contract,
                        kwargs=kwargs,
                    )
                    if tx_dict is None:
                        raise TxBuildError("Got empty transaction")

                    # Return transaction dict on dry-run
                    if dry_run:
                        return tx_dict

                    tx_signed = self.crypto.sign_transaction(transaction=tx_dict)
                    tx_digest = self.ledger_api.send_signed_transaction(
                        tx_signed=tx_signed,
                        raise_on_try=True,
                    )
                tx_receipt = self.ledger_api.api.eth.get_transaction_receipt(
                    cast(str, tx_digest)
                )
                if tx_receipt is not None:
                    return tx_receipt
            except RequestsConnectionError as e:
                raise RPCError("Cannot connect to the given RPC") from e
            except Exception as e:  # pylint: disable=broad-except
                error = str(e)
                if self._already_known(error):
                    already_known = True
                    continue  # pragma: nocover
                if not should_retry(error):
                    raise ChainInteractionError(error) from e
                if should_reprice(error):
                    print(f"Low gas error: {e}; Repricing the transaction...")
                    tx_dict = self._reprice(cast(Dict, tx_dict))
                    continue  # pragma: nocover
                if "Transaction with hash" in error and "not found" in error:
                    already_known = True
                    print(
                        f"Error getting transaction receipt: {e}; "
                        f"Will retry in {self.sleep}..."
                    )
                    time.sleep(self.sleep)
                    continue  # pragma: nocover

                if should_rebuild(error):
                    tx_dict = None

                tx_digest = None
                already_known = False
                print(
                    f"Error occured when interacting with chain: {e}; "
                    f"will retry in {self.sleep}..."
                )
                time.sleep(self.sleep)
        raise ChainTimeoutError("Timed out when waiting for transaction to go through")

    def process(
        self,
        event: str,
        receipt: Dict,
        contract: PublicId,
    ) -> Dict:
        """Process tx receipt."""
        return registry_contracts.get_contract(contract).get_events(
            ledger_api=self.ledger_api,
            contract_address=ContractConfigs.get(contract.name).contracts[
                self.chain_type
            ],
            receipt=receipt,
            event=event,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint/types/validator.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from packages.valory.connections.abci.gogoproto import (
    gogo_pb2 as gogoproto_dot_gogo__pb2,
)
from packages.valory.connections.abci.tendermint.crypto import (
    keys_pb2 as tendermint_dot_crypto_dot_keys__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n tendermint/types/validator.proto\x12\x10tendermint.types\x1a\x14gogoproto/gogo.proto\x1a\x1ctendermint/crypto/keys.proto"\x8a\x01\n\x0cValidatorSet\x12/\n\nvalidators\x18\x01 \x03(\x0b\x32\x1b.tendermint.types.Validator\x12-\n\x08proposer\x18\x02 \x01(\x0b\x32\x1b.tendermint.types.Validator\x12\x1a\n\x12total_voting_power\x18\x03 \x01(\x03"\x82\x01\n\tValidator\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\x0c\x12\x33\n\x07pub_key\x18\x02 \x01(\x0b\x32\x1c.tendermint.crypto.PublicKeyB\x04\xc8\xde\x1f\x00\x12\x14\n\x0cvoting_power\x18\x03 \x01(\x03\x12\x19\n\x11proposer_priority\x18\x04 \x01(\x03"V\n\x0fSimpleValidator\x12-\n\x07pub_key\x18\x01 \x01(\x0b\x32\x1c.tendermint.crypto.PublicKey\x12\x14\n\x0cvoting_power\x18\x02 \x01(\x03\x42\x39Z7github.com/tendermint/tendermint/proto/tendermint/typesb\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "tendermint.types.validator_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"Z7github.com/tendermint/tendermint/proto/tendermint/types"
    )
    _VALIDATOR.fields_by_name["pub_key"]._options = None
    _VALIDATOR.fields_by_name["pub_key"]._serialized_options = b"\310\336\037\000"
    _globals["_VALIDATORSET"]._serialized_start = 107
    _globals["_VALIDATORSET"]._serialized_end = 245
    _globals["_VALIDATOR"]._serialized_start = 248
    _globals["_VALIDATOR"]._serialized_end = 378
    _globals["_SIMPLEVALIDATOR"]._serialized_start = 380
    _globals["_SIMPLEVALIDATOR"]._serialized_end = 466
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the behaviours for the 'abci' skill."""

from abc import ABC
from collections import deque
from typing import Deque, Generator, Optional, Set, Type, cast

from packages.valory.contracts.squads_multisig.contract import (
    MultisigAccountType,
    SquadsMultisig,
    SquadsMultisigAuthorityIndex,
    TransactionStatus,
)
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.common import (
    RandomnessBehaviour,
    SelectKeeperBehaviour,
)
from packages.valory.skills.squads_transaction_settlement_abci.models import (
    SolanaTransactionSettlementParams,
)
from packages.valory.skills.squads_transaction_settlement_abci.payloads import (
    ApproveTxPayload,
    CreateTxPayload,
    ExecuteTxPayload,
    RandomnessPayload,
    SelectKeeperPayload,
    VerifyTxPayload,
)
from packages.valory.skills.squads_transaction_settlement_abci.rounds import (
    ApproveTxRound,
    CreateTxRandomnessRound,
    CreateTxRound,
    CreateTxSelectKeeperRound,
    ExecuteTxRandomnessRound,
    ExecuteTxRound,
    ExecuteTxSelectKeeperRound,
    SolanaTransactionSubmissionAbciApp,
    SynchronizedData,
    VerifyTxRound,
)


SOLANA = "solana"


class SolanaTransactionSettlementBaseBehaviour(BaseBehaviour, ABC):
    """Base behaviour for the common apps' skill."""

    @property
    def synchronized_data(self) -> SynchronizedData:
        """Return the synchronized data."""
        return cast(SynchronizedData, super().synchronized_data)

    @property
    def params(self) -> SolanaTransactionSettlementParams:
        """Return the params."""
        return cast(SolanaTransactionSettlementParams, super().params)

    def next_tx_index(self) -> Generator[None, None, Optional[int]]:
        """Get current tx index."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(SquadsMultisig.contract_id),
            contract_callable="next_tx_index",
            chain_id=SOLANA,
            ledger_id=SOLANA,
        )
        if response.performative != ContractApiMessage.Performative.STATE:
            return None
        return cast(int, response.state.body.pop("data"))

    def get_tx_pda(self, tx_index: int) -> Generator[None, None, Optional[str]]:
        """Create a new PDA for ms transaction."""
        response = yield from self.get_contract_api_response(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            contract_address=self.synchronized_data.safe_contract_address,
            contract_id=str(SquadsMultisig.contract_id),
            contract_callable="get_tx_pda",
            tx_index=tx_index,
            chain_id=SOLANA,
            ledger_id=SOLANA,
        )
        if response.performative != ContractApiMessage.Performative.STATE:
            return None
        return cast(str, response.state.body.pop("data"))

    @staticmethod
    def serialized_keepers(keepers: Deque[str], keeper_retries: int) -> str:
        """Get the keepers serialized."""
        if len(keepers) == 0:
            return ""
        keepers_ = "".join(keepers)
        keeper_retries_ = keeper_retries.to_bytes(32, "big").hex()
        concatenated = keeper_retries_ + keepers_

        return concatenated


class CreateTxRandomnessRoundRandomnessBehaviour(RandomnessBehaviour):
    """Retrieve randomness."""

    matching_round = CreateTxRandomnessRound
    payload_class = RandomnessPayload


class CreateTxSelectKeeperBehaviour(  # pylint: disable=too-many-ancestors
    SelectKeeperBehaviour, SolanaTransactionSettlementBaseBehaviour
):
    """Retrieve randomness."""

    matching_round = CreateTxSelectKeeperRound
    payload_class = SelectKeeperPayload

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = deque((self._select_keeper(),))
            payload = self.payload_class(
                self.context.agent_address, self.serialized_keepers(keepers, 1)
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class CreateTxBehaviour(SolanaTransactionSettlementBaseBehaviour):
    """Create a transaction with with given set of instructions."""

    matching_round = CreateTxRound

    def async_act(self) -> Generator:  # pylint: disable=inconsistent-return-statements
        """Create transaction."""
        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            next_tx_index = yield from self.next_tx_index()
            if next_tx_index is None:
                return None
            tx_pda = yield from self.get_tx_pda(tx_index=next_tx_index)
            if tx_pda is None:
                return None
            response = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(SquadsMultisig.contract_id),
                contract_callable="create_new_transaction_ix",
                authority_index=SquadsMultisigAuthorityIndex.VAULT.value,
                creator=self.context.agent_addresses["solana"],
                ixs=self.synchronized_data.most_voted_instruction_set,
                tx_pda=tx_pda,
                chain_id=SOLANA,
                ledger_id=SOLANA,
            )
            if response.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
                return None
            tx_digest, _ = yield from self.send_raw_transaction(
                transaction=response.raw_transaction,
                chain_id=SOLANA,
            )
            yield from self.sleep(5)

            self.context.logger.info(
                f"Created transaction with PDA {tx_pda} and hash={tx_digest}"
            )
            payload = CreateTxPayload(
                sender=self.context.agent_address,
                tx_pda=tx_pda,
                tx_digest=cast(str, tx_digest),
            )
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()
        self.set_done()


class ApproveTxBehaviour(SolanaTransactionSettlementBaseBehaviour):
    """Approve the transaction with given PDA."""

    matching_round = ApproveTxRound

    def async_act(self) -> Generator:  # pylint: disable=inconsistent-return-statements
        """Create transaction."""
        # TODO: Extend implementation
        # It might happen that the ApproveTxRound might not get an majority and
        # we will enter the approve tx round again. This means an agent can try to approve the
        # tx even if they already have done so. To prevent this, add a check which makes sure
        # if an agent has already approved/rejected the tx they don't make the transaction again
        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            response = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(SquadsMultisig.contract_id),
                contract_callable="approve_transaction_ix",
                tx_pda=self.synchronized_data.tx_pda,
                member=self.context.agent_addresses["solana"],
                chain_id=SOLANA,
                ledger_id=SOLANA,
            )
            if response.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
                return None
            yield from self.sleep(5)
            tx_digest, _ = yield from self.send_raw_transaction(
                transaction=response.raw_transaction,
                chain_id=SOLANA,
            )
            self.context.logger.info(
                f"Approved transaction with PDA {self.synchronized_data.tx_pda} and hash={tx_digest}"
            )
            payload = ApproveTxPayload(
                sender=self.context.agent_address,
                tx_pda=self.synchronized_data.tx_pda,
                tx_digest=cast(str, tx_digest),
            )
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()
        self.set_done()


class ExecuteTxRandomnessRoundRandomnessBehaviour(RandomnessBehaviour):
    """Retrieve randomness."""

    matching_round = ExecuteTxRandomnessRound
    payload_class = RandomnessPayload


class ExecuteTxSelectKeeperBehaviour(  # pylint: disable=too-many-ancestors
    SelectKeeperBehaviour, SolanaTransactionSettlementBaseBehaviour
):
    """Retrieve randomness."""

    matching_round = ExecuteTxSelectKeeperRound
    payload_class = SelectKeeperPayload

    def async_act(self) -> Generator:
        """Do the action."""

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            keepers = deque((self._select_keeper(),))
            payload = self.payload_class(
                self.context.agent_address, self.serialized_keepers(keepers, 1)
            )

        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()

        self.set_done()


class ExecuteTxBehaviour(SolanaTransactionSettlementBaseBehaviour):
    """Execute the transaction with given PDA."""

    matching_round = ExecuteTxRound

    def async_act(self) -> Generator:  # pylint: disable=inconsistent-return-statements
        """Create transaction."""
        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            response = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=self.synchronized_data.safe_contract_address,
                contract_id=str(SquadsMultisig.contract_id),
                contract_callable="execute_transaction_ix",
                tx_pda=self.synchronized_data.tx_pda,
                member=self.context.agent_addresses["solana"],
                chain_id=SOLANA,
                ledger_id=SOLANA,
            )
            if response.performative != ContractApiMessage.Performative.RAW_TRANSACTION:
                return None
            yield from self.sleep(5)

            tx_digest, _ = yield from self.send_raw_transaction(
                transaction=response.raw_transaction,
                chain_id=SOLANA,
            )
            self.context.logger.info(
                f"Executed transaction with PDA {self.synchronized_data.tx_pda} and hash={tx_digest}"
            )
            payload = ExecuteTxPayload(
                sender=self.context.agent_address,
                tx_pda=self.synchronized_data.tx_pda,
                tx_digest=cast(str, tx_digest),
            )
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()
        self.set_done()


class VerifyTxBehaviour(SolanaTransactionSettlementBaseBehaviour):
    """Execute the transaction with given PDA."""

    matching_round = VerifyTxRound
    _verification_retries: Optional[int] = None

    def async_act(self) -> Generator:  # pylint: disable=inconsistent-return-statements
        """Create transaction."""

        if self._verification_retries is None:
            self._verification_retries = 0

        if self._verification_retries > self.params.tx_verification_retries:
            payload = VerifyTxPayload(
                sender=self.context.agent_address,
                tx_pda=self.synchronized_data.tx_pda,
                verified=False,
            )
            with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
                yield from self.send_a2a_transaction(payload)
                yield from self.wait_until_round_end()
            return self.set_done()

        with self.context.benchmark_tool.measure(self.behaviour_id).local():
            response = yield from self.get_contract_api_response(
                performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
                contract_address=self.synchronized_data.tx_pda,
                contract_id=str(SquadsMultisig.contract_id),
                contract_callable="get_account_state",
                account_type=MultisigAccountType.MS_TRANSACTION.value,
                chain_id=SOLANA,
                ledger_id=SOLANA,
            )
            if response.performative != ContractApiMessage.Performative.STATE:
                return None
            account_state = response.state.body
            if TransactionStatus(account_state["status"]) != TransactionStatus.Executed:
                yield from self.sleep(self.params.tx_verification_sleep)
                self._verification_retries += 1
                return None
            self._verification_retries = 0
            payload = VerifyTxPayload(
                sender=self.context.agent_address,
                tx_pda=self.synchronized_data.tx_pda,
                verified=True,
            )
        with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
            yield from self.send_a2a_transaction(payload)
            yield from self.wait_until_round_end()
        self.set_done()


class SolanaTransactionSettlementRoundBehaviour(AbstractRoundBehaviour):
    """This behaviour manages the consensus stages for the basic transaction settlement."""

    initial_behaviour_cls = CreateTxRandomnessRoundRandomnessBehaviour
    abci_app_cls = SolanaTransactionSubmissionAbciApp
    behaviours: Set[Type[BaseBehaviour]] = {
        CreateTxRandomnessRoundRandomnessBehaviour,  # type: ignore
        CreateTxSelectKeeperBehaviour,  # type: ignore
        CreateTxBehaviour,  # type: ignore
        ApproveTxBehaviour,  # type: ignore
        ExecuteTxRandomnessRoundRandomnessBehaviour,  # type: ignore
        ExecuteTxSelectKeeperBehaviour,  # type: ignore
        ExecuteTxBehaviour,  # type: ignore
        VerifyTxBehaviour,  # type: ignore
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""ETH encoder."""

import typing as t

from Crypto.Hash import keccak  # nosec
from eth_abi.abi import default_codec  # nosec
from eth_utils import decode_hex


def encode(typ: t.Any, arg: t.Any) -> bytes:
    """Encode by type."""
    encoder = default_codec._registry.get_encoder(  # pylint: disable=protected-access
        typ
    )
    return encoder(arg)


def to_string(value: t.Union[bytes, str, int]) -> bytes:
    """Convert to byte string."""
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return bytes(value, "utf-8")
    if isinstance(value, int):
        return bytes(str(value), "utf-8")
    raise ValueError("Invalid data")


def sha3_256(x: bytes) -> bytes:
    """Calculate SHA3-256 hash."""
    return keccak.new(digest_bits=256, data=x).digest()


def sha3(seed: t.Union[bytes, str, int]) -> bytes:
    """Calculate SHA3-256 hash."""
    return sha3_256(to_string(seed))


def scan_bin(v: str) -> bytes:
    """Scan bytes."""
    if v[:2] in ("0x", b"0x"):
        return decode_hex(v[2:])
    return decode_hex(v)


def create_struct_definition(name: str, schema: t.List[t.Dict[str, str]]) -> str:
    """Create method struction defintion."""
    schema_types = [
        (schema_type["type"] + " " + schema_type["name"]) for schema_type in schema
    ]
    return name + "(" + ",".join(schema_types) + ")"


def find_dependencies(
    name: str, types: t.Dict[str, t.Any], dependencies: t.Set
) -> None:
    """Find dependencies."""
    if name in dependencies:
        return
    schema = types.get(name)
    if not schema:
        return
    dependencies.add(name)
    for schema_type in schema:
        find_dependencies(schema_type["type"], types, dependencies)


def create_schema(name: str, types: t.Dict) -> str:
    """Create schema."""
    array_start = name.find("[")
    clean_name = name if array_start < 0 else name[:array_start]
    dependencies: t.Set = set()
    find_dependencies(clean_name, types, dependencies)
    dependencies.discard(clean_name)
    dependency_definitions = [
        create_struct_definition(dependency, types[dependency])
        for dependency in sorted(dependencies)
        if types.get(dependency)
    ]
    return create_struct_definition(clean_name, types[clean_name]) + "".join(
        dependency_definitions
    )


def create_schema_hash(name: str, types: t.Dict) -> bytes:
    """Create schema hash."""
    return encode("bytes32", sha3(create_schema(name, types)))


def encode_value(data_type: str, value: t.Any, types: t.Dict) -> bytes:
    """Encode value."""
    if data_type == "string":
        return encode("bytes32", sha3(value))
    if data_type == "bytes":
        return encode("bytes32", sha3(scan_bin(value)))
    if types.get(data_type):
        return encode("bytes32", sha3(encode_data(data_type, value, types)))
    if data_type.endswith("]"):
        arrayType = data_type[: data_type.index("[")]
        return encode(
            "bytes32",
            sha3(
                b"".join(
                    [encode_data(arrayType, arrayValue, types) for arrayValue in value]
                )
            ),
        )
    return encode(data_type, value)


def encode_data(name: str, data: t.Dict[str, t.Dict[str, str]], types: t.Dict) -> bytes:
    """Encode data."""
    return create_schema_hash(name, types) + b"".join(
        [
            encode_value(schema_type["type"], data[schema_type["name"]], types)
            for schema_type in types[name]
        ]
    )


def create_struct_hash(
    name: str, data: t.Dict[str, t.Dict[str, str]], types: t.Dict
) -> bytes:
    """Create struct hash."""
    return sha3(encode_data(name, data, types))


def encode_typed_data(data: t.Dict[str, t.Any]) -> bytes:
    """Encode typed data."""
    types = t.cast(t.Dict, data.get("types"))
    primary_type = t.cast(str, data.get("primaryType"))
    domain = t.cast(t.Dict[str, t.Dict[str, str]], data.get("domain"))
    message = t.cast(t.Dict[str, t.Any], data.get("message"))
    domain_hash = create_struct_hash("EIP712Domain", domain, types)
    message_hash = create_struct_hash(primary_type, message, types)
    return sha3(bytes.fromhex("19") + bytes.fromhex("01") + domain_hash + message_hash)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2025 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Docker-compose Deployment Generator."""

import ipaddress
import os
from pathlib import Path
from typing import Dict, List, Optional, Set, cast

import yaml
from aea.configurations.constants import (
    DEFAULT_LEDGER,
    LEDGER,
    PRIVATE_KEY,
    PRIVATE_KEY_PATH_SCHEMA,
)
from docker import DockerClient, from_env
from docker.constants import DEFAULT_NPIPE, IS_WINDOWS_PLATFORM
from docker.errors import DockerException

from autonomy.configurations.constants import DEFAULT_SERVICE_CONFIG_FILE
from autonomy.constants import (
    ACN_IMAGE_NAME,
    ACN_IMAGE_VERSION,
    DEVELOPMENT_IMAGE,
    DOCKER_COMPOSE_YAML,
    HARDHAT_IMAGE_NAME,
    HARDHAT_IMAGE_VERSION,
    OAR_IMAGE,
    TENDERMINT_IMAGE_NAME,
    TENDERMINT_IMAGE_VERSION,
)
from autonomy.deploy.base import (
    BaseDeploymentGenerator,
    DEFAULT_RESOURCE_VALUES,
    Resources,
    tm_write_to_log,
)
from autonomy.deploy.constants import (
    DEFAULT_ENCODING,
    DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA,
    DEPLOYMENT_KEY_DIRECTORY,
    INFO,
)
from autonomy.deploy.generators.docker_compose.templates import (
    ABCI_NODE_TEMPLATE,
    ACN_NODE_TEMPLATE,
    DOCKER_COMPOSE_TEMPLATE,
    HARDHAT_NODE_TEMPLATE,
    PORTS,
    PORT_MAPPING_CONFIG,
    TENDERMINT_CONFIG_TEMPLATE,
    TENDERMINT_NODE_TEMPLATE,
)


NETWORK_ADDRESS_OFFSET = 1
BASE_SUBNET = cast(ipaddress.IPv4Network, ipaddress.ip_network("192.167.11.0/24"))
SUBNET_OVERFLOW = 256

DEFAULT_PACKAGES_PATH = Path.cwd().absolute() / "packages"
DEFAULT_OPEN_AEA_DIR: Path = Path.home().absolute() / "open-aea"


def get_docker_client() -> DockerClient:
    """Load docker client."""
    try:
        return from_env()
    except DockerException:
        return DockerClient(
            DEFAULT_NPIPE
            if IS_WINDOWS_PLATFORM
            else f"unix://{Path.home()}/.docker/desktop/docker.sock"
        )


def build_tendermint_node_config(  # pylint: disable=too-many-arguments
    node_id: int,
    container_name: str,
    abci_node: str,
    network_name: str,
    network_address: str,
    dev_mode: bool = False,
    log_level: str = INFO,
    tendermint_ports: Optional[Dict[int, int]] = None,
) -> str:
    """Build tendermint node config for docker compose."""
    config = TENDERMINT_NODE_TEMPLATE.format(
        node_id=node_id,
        container_name=container_name,
        abci_node=abci_node,
        network_address=network_address,
        localnet_port_range=node_id,
        log_level=log_level,
        tendermint_image_name=TENDERMINT_IMAGE_NAME,
        tendermint_image_version=TENDERMINT_IMAGE_VERSION,
        network_name=network_name,
        write_to_log=str(tm_write_to_log()).lower(),
        user=os.environ.get("UID", "1000"),
    )

    if dev_mode:
        config += "      - ./persistent_data/tm_state:/tm_state:Z"
        config = config.replace("DEV_MODE=0", "DEV_MODE=1")

    if tendermint_ports is not None:
        port_mappings = map(
            lambda x: PORT_MAPPING_CONFIG.format(host_port=x[0], container_port=x[1]),
            tendermint_ports.items(),
        )
        port_config = "\n".join([PORTS, *port_mappings])
        config += port_config
        config += "\n"

    return config


def to_env_file(agent_vars: Dict, node_id: int, build_dir: Path) -> None:
    """Create a env file under the `agent_build` folder."""
    agent_vars["PYTHONHASHSEED"] = 0
    agent_vars["LOG_FILE"] = f"/logs/aea_{node_id}.txt"
    env_file_path = build_dir / f"agent_{node_id}.env"
    with open(env_file_path, "w", encoding=DEFAULT_ENCODING) as env_file:
        for key, value in agent_vars.items():
            env_file.write(f"{key}={value}\n")


def build_agent_config(  # pylint: disable=too-many-arguments,too-many-locals
    node_id: int,
    build_dir: Path,
    container_name: str,
    agent_vars: Dict,
    runtime_image: str,
    network_name: str,
    network_address: str,
    dev_mode: bool = False,
    package_dir: Optional[Path] = None,
    open_aea_dir: Optional[Path] = None,
    agent_ports: Optional[Dict[int, int]] = None,
    extra_volumes: Optional[Dict[str, str]] = None,
    resources: Optional[Resources] = None,
) -> str:
    """Build agent config."""
    resources = resources if resources is not None else DEFAULT_RESOURCE_VALUES
    to_env_file(agent_vars, node_id, build_dir)
    config = ABCI_NODE_TEMPLATE.format(
        node_id=node_id,
        container_name=container_name,
        network_address=network_address,
        runtime_image=runtime_image,
        env_file=f"agent_{node_id}.env",
        network_name=network_name,
        agent_memory_request=resources["agent"]["requested"]["memory"],
        agent_cpu_limit=resources["agent"]["limit"]["cpu"],
        agent_memory_limit=resources["agent"]["limit"]["memory"],
    )

    if dev_mode:
        config += "      - ./persistent_data/benchmarks:/benchmarks:Z\n"
        config += "      - ./persistent_data/venvs:/root/.local/share/virtualenvs:Z\n"
        config += f"      - {package_dir}:/root/packages:rw\n"
        config += f"      - {open_aea_dir}:/open-aea\n"

    if extra_volumes is not None:
        for host_dir, container_dir in extra_volumes.items():
            config += f"      - {host_dir}:{container_dir}:Z\n"
            Path(host_dir).resolve().mkdir(exist_ok=True, parents=True)

    if agent_ports is not None:
        port_mappings = map(
            lambda x: PORT_MAPPING_CONFIG.format(host_port=x[0], container_port=x[1]),
            agent_ports.items(),
        )
        port_config = "\n".join([PORTS, *port_mappings])
        config += port_config
        config += "\n"

    return config


class Network:
    """Class to represent network of the service."""

    def __init__(
        self,
        name: str,
        base: ipaddress.IPv4Network = BASE_SUBNET,
        used_subnets: Optional[Set[str]] = None,
    ) -> None:
        """Initialize."""
        self.name = name
        self.base = base
        self.used_subnets = set() if used_subnets is None else used_subnets
        self.subnet = self.build()

        self._address_offeset = NETWORK_ADDRESS_OFFSET

    @staticmethod
    def next_subnet(subnet: ipaddress.IPv4Network) -> ipaddress.IPv4Network:
        """Calculat next available subnet."""
        new_address = subnet.network_address + SUBNET_OVERFLOW
        return cast(
            ipaddress.IPv4Network,
            ipaddress.ip_network(f"{new_address}/{subnet.prefixlen}"),
        )

    def build(
        self,
    ) -> ipaddress.IPv4Network:
        """Initialize network params."""
        docker = get_docker_client()
        used_subnets = self.used_subnets
        for network in docker.networks.list():
            # Network already exists
            if f"abci_build_{self.name}" == network.attrs["Name"]:
                config, *_ = network.attrs["IPAM"]["Config"]
                return cast(
                    ipaddress.IPv4Network, ipaddress.ip_network(config["Subnet"])
                )
            if network.attrs["IPAM"]["Config"] is None:
                continue
            for config in network.attrs["IPAM"]["Config"]:
                used_subnets.add(config["Subnet"])

        subnet = self.base
        while str(subnet) in used_subnets:
            subnet = self.next_subnet(subnet=subnet)
        return subnet

    @property
    def next_address(self) -> str:
        """Returns the next IP address string."""
        self._address_offeset += 1
        return str(self.subnet.network_address + self._address_offeset)


class DockerComposeGenerator(BaseDeploymentGenerator):
    """Class to automate the generation of Deployments."""

    output_name: str = DOCKER_COMPOSE_YAML
    deployment_type: str = "docker-compose"

    def _find_occupied_networks(self, network_name: str) -> Set[str]:
        """Find occupied networks on the host system."""
        used_ports: Dict[str, str] = {}  # find occupied ports
        for config_name in ("agent", "tendermint"):
            for port_mapping in (
                self.service_builder.service.deployment_config.get(config_name, {})
                .get("ports", {})
                .values()
            ):
                for host_port, _ in port_mapping.items():
                    if host_port in used_ports:
                        print(
                            f"WARNING: Port {host_port} is already used by {used_ports[host_port]}."
                        )
                    else:
                        used_ports[str(host_port)] = DEFAULT_SERVICE_CONFIG_FILE

        used_networks = set()  # find occupied networks
        for build in Path.cwd().glob("abci_build_*"):
            if not (build / DOCKER_COMPOSE_YAML).exists():
                continue

            compose_config = yaml.safe_load((build / DOCKER_COMPOSE_YAML).read_text())
            for subnet in (
                compose_config.get("networks", {})
                .get(network_name, {})
                .get("ipam", {})
                .get("config", [])
            ):
                used_networks.add(subnet["subnet"])

            services = compose_config.get("services", {})
            for name, service in services.items():
                for port_mapping_str in service.get("ports", []):
                    port_mapping = port_mapping_str.split(":")
                    if len(port_mapping) != 2:
                        continue

                    host_port = port_mapping[0]
                    if host_port in used_ports:
                        print(
                            f"WARNING: Port {host_port} is already used by {used_ports[host_port]}."
                            f" Please adjust the port in {build / DOCKER_COMPOSE_YAML} manually.",
                        )
                    else:
                        used_ports[host_port] = name

        return used_networks

    def generate_config_tendermint(self) -> "DockerComposeGenerator":
        """Generate the command to configure tendermint testnet."""
        if not self.use_tm_testnet_setup:
            return self

        hosts = " ".join(
            [
                "--hostname=" + self.service_builder.get_tm_container_name(index=k)
                for k in range(self.service_builder.service.number_of_agents)
            ]
        )
        self.tendermint_job_config = TENDERMINT_CONFIG_TEMPLATE.format(
            validators=self.service_builder.service.number_of_agents,
            hosts=hosts,
            user=os.environ.get("UID", "1000"),
        )
        client = get_docker_client()
        image = f"{TENDERMINT_IMAGE_NAME}:{TENDERMINT_IMAGE_VERSION}"

        run_log = client.containers.run(
            image=image,
            volumes={f"{self.build_dir}/nodes": {"bind": "/tendermint", "mode": "z"}},
            entrypoint=self.tendermint_job_config,
        )
        print(run_log.decode())

        return self

    def generate(
        self,
        image_version: Optional[str] = None,
        use_hardhat: bool = False,
        use_acn: bool = False,
    ) -> "DockerComposeGenerator":
        """Generate the new configuration."""
        network_name = f"service_{self.service_builder.service.name}_localnet"
        used_networks = self._find_occupied_networks(network_name)
        network = Network(name=network_name, used_subnets=used_networks)
        image_version = image_version or self.service_builder.service.agent.hash
        if self.dev_mode:
            runtime_image = DEVELOPMENT_IMAGE
        else:
            runtime_image = OAR_IMAGE.format(
                image_author=self.image_author,
                agent=self.service_builder.service.agent.name,
                version=image_version,
            )

        agent_vars = self.service_builder.generate_agents()
        agents = "".join(
            [
                build_agent_config(
                    node_id=i,
                    container_name=self.service_builder.get_abci_container_name(
                        index=i
                    ),
                    build_dir=self.build_dir,
                    runtime_image=runtime_image,
                    agent_vars=agent_vars[i],
                    dev_mode=self.dev_mode,
                    package_dir=self.packages_dir,
                    open_aea_dir=self.open_aea_dir,
                    agent_ports=(
                        self.service_builder.service.deployment_config.get("agent", {})
                        .get("ports", {})
                        .get(i)
                    ),
                    network_name=network_name,
                    network_address=network.next_address,
                    resources=self.resources,
                    extra_volumes=self.service_builder.service.deployment_config.get(
                        "agent", {}
                    ).get("volumes"),
                )
                for i in range(self.service_builder.service.number_of_agents)
            ]
        )
        tendermint_nodes = "".join(
            [
                build_tendermint_node_config(
                    node_id=i,
                    container_name=self.service_builder.get_tm_container_name(index=i),
                    abci_node=self.service_builder.get_abci_container_name(index=i),
                    dev_mode=self.dev_mode,
                    log_level=self.service_builder.log_level,
                    tendermint_ports=(
                        self.service_builder.service.deployment_config.get(
                            "tendermint", {}
                        )
                        .get("ports", {})
                        .get(i)
                    ),
                    network_name=network_name,
                    network_address=network.next_address,
                )
                for i in range(self.service_builder.service.number_of_agents)
            ]
        )

        hardhat_node = ""
        if use_hardhat:
            hardhat_node = HARDHAT_NODE_TEMPLATE.format(
                hardhat_image_name=HARDHAT_IMAGE_NAME,
                hardhat_image_version=HARDHAT_IMAGE_VERSION,
                network_name=network_name,
                network_address=network.next_address,
            )

        acn_node = ""
        if use_acn:
            acn_node = ACN_NODE_TEMPLATE.format(
                acn_image_name=ACN_IMAGE_NAME,
                acn_image_version=ACN_IMAGE_VERSION,
                network_name=network_name,
                network_address=network.next_address,
            )

        self.output = DOCKER_COMPOSE_TEMPLATE.format(
            abci_nodes=agents,
            tendermint_nodes=tendermint_nodes,
            hardhat_node=hardhat_node,
            acn_node=acn_node,
            network_name=network_name,
            subnet=str(network.subnet),
        )

        return self

    def _populate_keys(self) -> None:
        """Populate the keys directory"""
        keys_dir = self.build_dir / DEPLOYMENT_KEY_DIRECTORY
        for x in range(self.service_builder.service.number_of_agents):
            path = keys_dir / DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA.format(agent_n=x)
            ledger = cast(List[Dict[str, str]], self.service_builder.keys)[x].get(
                LEDGER, DEFAULT_LEDGER
            )
            key = cast(List[Dict[str, str]], self.service_builder.keys)[x][PRIVATE_KEY]
            keys_file = path / PRIVATE_KEY_PATH_SCHEMA.format(ledger)
            path.mkdir(parents=True)
            with keys_file.open(mode="w", encoding=DEFAULT_ENCODING) as f:
                f.write(key)

    def _populate_keys_multiledger(self) -> None:
        """Populate the keys directory with multiple set of keys"""
        keys_dir = self.build_dir / DEPLOYMENT_KEY_DIRECTORY
        for x in range(self.service_builder.service.number_of_agents):
            path = keys_dir / DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA.format(agent_n=x)
            path.mkdir(exist_ok=True)
            for keypair in cast(List[List[Dict[str, str]]], self.service_builder.keys)[
                x
            ]:
                ledger = keypair.get(LEDGER, DEFAULT_LEDGER)
                key = keypair[PRIVATE_KEY]
                keys_file = path / PRIVATE_KEY_PATH_SCHEMA.format(ledger)
                with keys_file.open(mode="w", encoding=DEFAULT_ENCODING) as f:
                    f.write(key)

    def populate_private_keys(
        self,
    ) -> "DockerComposeGenerator":
        """Populate the private keys to the build directory for docker-compose mapping."""
        if self.service_builder.multiledger:
            self._populate_keys_multiledger()
        else:
            self._populate_keys()
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/offend_abci skill's behaviours."""

from copy import deepcopy
from pathlib import Path
from typing import Any, cast

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    AbciAppDB,
    BaseSynchronizedData,
    OffenceStatus,
)
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.offend_abci.behaviours import OffendBehaviour
from packages.valory.skills.offend_abci.rounds import Event, FinishedOffendRound


PACKAGE_DIR = Path(__file__).parents[1]


FIRST_PARTICIPANT = "a"
OTHER_PARTICIPANTS = ("c", "b", "d")
ALL_PARTICIPANTS = OTHER_PARTICIPANTS + (FIRST_PARTICIPANT,)
LIGHT_OFFENCE_UNIT_AMOUNT = 1
SERIOUS_OFFENCE_UNIT_AMOUNT = 2
INITIAL_OFFENCE_STATUS = {
    participant: OffenceStatus() for participant in ALL_PARTICIPANTS
}


class TestOffendBehaviour(FSMBehaviourBaseCase):
    """Test the `OffendBehaviour`."""

    path_to_skill = PACKAGE_DIR
    synchronized_data: BaseSynchronizedData

    @property
    def current_behaviour(self) -> OffendBehaviour:
        """Get the current behaviour."""
        return cast(OffendBehaviour, self.behaviour.current_behaviour)

    def setup(self, **kwargs: Any) -> None:
        """
        Set up the test method.

        Called each time before a test method is called.

        :param kwargs: the keyword arguments passed to _prepare_skill
        """
        super().setup(**kwargs)
        self.synchronized_data = BaseSynchronizedData(
            AbciAppDB(
                setup_data=AbciAppDB.data_to_lists(
                    {"all_participants": ALL_PARTICIPANTS}
                )
            )
        )
        self.fast_forward_to_behaviour(
            self.behaviour,
            OffendBehaviour.auto_behaviour_id(),
            self.synchronized_data,
        )
        assert self.current_behaviour is not None
        assert (
            self.current_behaviour.behaviour_id == OffendBehaviour.auto_behaviour_id()
        )

        self.current_behaviour.context.state.round_sequence.offence_status = deepcopy(
            INITIAL_OFFENCE_STATUS
        )

        self.current_behaviour.params.__dict__["_frozen"] = False
        self.current_behaviour.params.suspected = True
        self.current_behaviour.params.__dict__["_frozen"] = True

    @pytest.mark.parametrize(
        (
            "suspected",
            "validator_downtime",
            "invalid_payload",
            "blacklisted",
            "num_unknown",
            "num_double_signed",
            "num_light_client_attack",
            "expected_slash_amount",
        ),
        (
            (False, False, False, False, 0, 0, 0, 0),
            (True, False, False, False, 0, 0, 0, 1),
            (False, True, False, False, 0, 0, 0, 1),
            (False, False, True, False, 0, 0, 0, 1),
            (False, False, False, True, 0, 0, 0, 1),
            (False, False, False, False, 1, 0, 0, 2),
            (False, False, False, False, 0, 1, 0, 2),
            (False, False, False, False, 0, 0, 1, 2),
            (False, False, False, False, 0, 2, 1, 6),
            (False, True, False, True, 5, 2, 1, 18),
            (True, True, True, True, 5, 2, 1, 20),
        ),
    )
    def test_updated_status(  # pylint: disable=too-many-arguments
        self,
        suspected: bool,
        validator_downtime: bool,
        invalid_payload: bool,
        blacklisted: bool,
        num_unknown: int,
        num_double_signed: int,
        num_light_client_attack: int,
        expected_slash_amount: str,
    ) -> None:
        """Test the `_updated_status` method."""
        self.current_behaviour.params.__dict__["_frozen"] = False
        self.current_behaviour.params.suspected = suspected
        self.current_behaviour.params.validator_downtime = validator_downtime
        self.current_behaviour.params.invalid_payload = invalid_payload
        self.current_behaviour.params.blacklisted = blacklisted
        self.current_behaviour.params.num_unknown = num_unknown
        self.current_behaviour.params.num_double_signed = num_double_signed
        self.current_behaviour.params.num_light_client_attack = num_light_client_attack
        self.current_behaviour.params.__dict__["_frozen"] = True

        status = (
            self.current_behaviour._updated_status()  # pylint: disable=protected-access
        )

        actual_slash_amount = status[FIRST_PARTICIPANT].slash_amount(
            LIGHT_OFFENCE_UNIT_AMOUNT, SERIOUS_OFFENCE_UNIT_AMOUNT
        )
        assert actual_slash_amount == expected_slash_amount

        for participant in OTHER_PARTICIPANTS:
            actual_slash_amount = status[participant].slash_amount(
                LIGHT_OFFENCE_UNIT_AMOUNT, SERIOUS_OFFENCE_UNIT_AMOUNT
            )
            assert actual_slash_amount == 0

    def test_offend_act(self) -> None:
        """Test offend behaviour's `async_act` method."""
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        status = self.current_behaviour.shared_state.round_sequence.offence_status
        assert status == INITIAL_OFFENCE_STATUS
        self._test_done_flag_set()
        self.end_round(Event.DONE)
        expected_id = f"degenerate_behaviour_{FinishedOffendRound.auto_round_id()}"
        assert self.current_behaviour.behaviour_id == expected_id



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""On-chain interaction helpers."""

import binascii
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, cast

import click
from aea.configurations.base import PackageConfiguration
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.configurations.loader import load_configuration_object
from aea.crypto.base import Crypto, LedgerApi
from aea.crypto.registries import crypto_registry, ledger_apis_registry
from aea.helpers.base import IPFSHash
from requests.exceptions import ConnectionError as RequestConnectionError
from texttable import Texttable

from autonomy.chain.base import ServiceState, UnitType
from autonomy.chain.config import (
    ChainConfigs,
    ChainType,
    ContractConfig,
    ContractConfigs,
)
from autonomy.chain.constants import (
    AGENT_REGISTRY_CONTRACT,
    COMPONENT_REGISTRY_CONTRACT,
    SERVICE_REGISTRY_CONTRACT,
    SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT,
)
from autonomy.chain.exceptions import ChainInteractionError
from autonomy.chain.metadata import NFTHashOrPath, publish_metadata
from autonomy.chain.mint import DEFAULT_NFT_IMAGE_HASH, MintManager
from autonomy.chain.service import (
    ServiceManager,
    approve_erc20_usage,
    get_activate_registration_amount,
    get_agent_instances,
    get_service_info,
    get_token_deposit_amount,
    is_service_token_secured,
)
from autonomy.chain.subgraph.client import SubgraphClient
from autonomy.chain.utils import (
    is_service_manager_token_compatible_chain,
    resolve_component_id,
)
from autonomy.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS, Service


try:
    from aea_ledger_ethereum.ethereum import EthereumApi, EthereumCrypto

    ETHEREUM_PLUGIN_INSTALLED = True
except ImportError:  # pragma: nocover
    ETHEREUM_PLUGIN_INSTALLED = False


class OnChainHelper:  # pylint: disable=too-few-public-methods
    """On-chain interaction helper."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        chain_type: ChainType,
        key: Optional[Path] = None,
        password: Optional[str] = None,
        hwi: bool = False,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        sleep: Optional[float] = None,
        dry_run: bool = False,
    ) -> None:
        """Initialize object."""
        if key is None and not hwi:
            raise click.ClickException(
                "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
            )

        self.chain_type = chain_type
        self.ledger_api, self.crypto = self.get_ledger_and_crypto_objects(
            chain_type=chain_type,
            key=key,
            password=password,
            hwi=hwi,
        )
        self.timeout = timeout
        self.retries = retries
        self.sleep = sleep
        self.dry_run = dry_run

    @staticmethod
    def load_hwi_plugin() -> Type[LedgerApi]:  # pragma: nocover
        """Load HWI Plugin."""
        try:
            from aea_ledger_ethereum_hwi.hwi import (  # pylint: disable=import-outside-toplevel
                EthereumHWIApi,
            )

            return EthereumHWIApi
        except ImportError as e:
            raise click.ClickException(
                "Hardware wallet plugin not installed, "
                "Run `pip3 install open-aea-ledger-ethereum-hwi` to install the plugin"
            ) from e
        except TypeError as e:
            raise click.ClickException(
                'Protobuf compatibility error; Please export PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION="python" '
                "to use the hardware wallet without any issues"
            ) from e

    @staticmethod
    def load_crypto(
        file: Path,
        password: Optional[str] = None,
    ) -> Crypto:
        """Load crypto object."""
        try:
            return EthereumCrypto(
                private_key_path=file,
                password=password,
            )
        except (binascii.Error, ValueError) as e:
            raise click.ClickException(
                "Cannot load private key for following possible reasons\n"
                "- Wrong key format\n"
                "- Wrong key length\n"
                "- Trailing spaces or new line characters"
            ) from e

    @classmethod
    def get_ledger_and_crypto_objects(
        cls,
        chain_type: ChainType,
        key: Optional[Path] = None,
        password: Optional[str] = None,
        hwi: bool = False,
    ) -> Tuple[LedgerApi, Crypto]:
        """Create ledger_api and crypto objects"""
        chain_config = ChainConfigs.get(chain_type=chain_type)
        identifier = EthereumApi.identifier

        if chain_config.rpc is None:
            raise click.ClickException(
                f"RPC URL cannot be `None`, "
                f"Please set the environment variable for {chain_type.value} chain "
                f"using `{ChainConfigs.get_rpc_env_var(chain_type)}` environment variable"
            )

        if hwi:
            EthereumHWIApi = cls.load_hwi_plugin()
            identifier = EthereumHWIApi.identifier

        if not hwi and not ETHEREUM_PLUGIN_INSTALLED:  # pragma: nocover
            raise click.ClickException(
                "Ethereum ledger plugin not installed, "
                "Run `pip3 install open-aea-ledger-ethereum` to install the plugin"
            )

        if key is None:
            crypto = crypto_registry.make(identifier)
        else:
            crypto = cls.load_crypto(
                file=key,
                password=password,
            )

        ledger_api = ledger_apis_registry.make(
            identifier,
            **{
                "address": chain_config.rpc,
                "chain_id": chain_config.chain_id,
                "is_gas_estimation_enabled": True,
            },
        )

        if hwi:
            # Setting the `LedgerApi.identifier` to `ethereum` for both ledger and
            # hardware plugin to interact with the contract. If we use `ethereum_hwi`
            # as the ledger identifier the contracts will need ABI configuration for
            # the `ethereum_hwi` identifier which means we will have to define hardware
            # wallet as the dependency for contract but the hardware wallet plugin
            # is meant to be used for CLI tools only so we set the identifier to
            # `ethereum` for both ledger and hardware wallet plugin
            ledger_api.identifier = EthereumApi.identifier

        try:
            ledger_api.api.eth.default_account = crypto.address
        except Exception as e:  # pragma: nocover
            raise click.ClickException(str(e))

        return ledger_api, crypto

    def check_required_enviroment_variables(
        self, configs: Tuple[ContractConfig, ...]
    ) -> None:
        """Check for required enviroment variables when working with the custom chain."""
        if self.chain_type != ChainType.CUSTOM:
            return
        missing = []
        for config in configs:
            if config.contracts[self.chain_type] is None:
                missing.append(config)

        if len(missing) == 0:
            return

        error = "Addresses for following contracts are None, please set them using their respective environment variables\n"
        for config in missing:
            error += f"- Set `{config.name}` address using `CUSTOM_{config.name.upper()}_ADDRESS`\n"
        raise click.ClickException(error[:-1])


class MintHelper(OnChainHelper):  # pylint: disable=too-many-instance-attributes
    """Mint helper."""

    package_path: Path
    package_type: PackageType
    package_configuration: PackageConfiguration
    old_metadata: Dict[str, Any]
    nft: NFTHashOrPath
    dependencies: List[int]
    agent_id: int
    metadata_hash: str
    metadata_string: str
    token_id: Optional[int]

    def __init__(  # pylint: disable=too-many-arguments
        self,
        chain_type: ChainType,
        key: Optional[Path] = None,
        password: Optional[str] = None,
        hwi: bool = False,
        update_token: Optional[int] = None,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        sleep: Optional[float] = None,
        subgraph: Optional[str] = None,
        dry_run: bool = False,
    ) -> None:
        """Initialize object."""
        super().__init__(
            chain_type,
            key,
            password,
            hwi,
            timeout=timeout,
            retries=retries,
            sleep=sleep,
            dry_run=dry_run,
        )
        self.update_token = update_token
        self.manager = MintManager(
            ledger_api=self.ledger_api,
            crypto=self.crypto,
            chain_type=chain_type,
            timeout=timeout,
            retries=retries,
            sleep=sleep,
            dry_run=self.dry_run,
        )
        self.subgraph = SubgraphClient(url=subgraph)

    def load_package_configuration(
        self,
        package_path: Path,
        package_type: PackageType,
    ) -> "MintHelper":
        """Load package configuration."""
        try:
            self.package_path = package_path
            self.package_type = package_type
            self.package_configuration = load_configuration_object(
                package_type=package_type,
                directory=package_path,
                package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
            )
            return self
        except FileNotFoundError as e:  # pragma: nocover
            raise click.ClickException(
                f"Cannot find configuration file for {package_type}"
            ) from e

    def load_metadata(self) -> "MintHelper":  # pragma: nocover
        """Load metadata when updating a mint."""
        if self.update_token is None:
            return self

        if self.package_type == PackageType.SERVICE:
            is_service = True
            is_agent = False
            contract_address = ContractConfigs.get(
                SERVICE_REGISTRY_CONTRACT.name
            ).contracts[self.chain_type]
            self.check_required_enviroment_variables(
                configs=(ContractConfigs.service_registry,)
            )
        elif self.package_type == PackageType.AGENT:
            is_service = False
            is_agent = True
            contract_address = ContractConfigs.get(
                AGENT_REGISTRY_CONTRACT.name
            ).contracts[self.chain_type]
            self.check_required_enviroment_variables(
                configs=(ContractConfigs.agent_registry,)
            )
        else:
            is_service = False
            is_agent = False
            contract_address = ContractConfigs.get(
                COMPONENT_REGISTRY_CONTRACT.name
            ).contracts[self.chain_type]
            self.check_required_enviroment_variables(
                configs=(ContractConfigs.component_registry,)
            )

        self.old_metadata = resolve_component_id(
            ledger_api=self.ledger_api,
            contract_address=contract_address,
            is_agent=is_agent,
            is_service=is_service,
            token_id=self.update_token,
        )

        return self

    def verify_nft(self, nft: Optional[NFTHashOrPath] = None) -> "MintHelper":
        """Verify NFT image."""

        # If new NFT hash is not provided, check previous mint for NFT hash
        if self.update_token is not None and nft is None:
            image_uri = self.old_metadata.get("image", None)
            if image_uri is not None:
                nft = IPFSHash(image_uri.replace("ipfs://", ""))

        # NFTs are required for minting components on a local chain deployement
        if self.chain_type == ChainType.LOCAL and nft is None:
            nft = IPFSHash(DEFAULT_NFT_IMAGE_HASH)

        if self.chain_type != ChainType.LOCAL and nft is None:
            raise click.ClickException(
                f"Please provide hash for NFT image to mint component on `{self.chain_type.value}` chain"
            )

        self.nft = nft
        return self

    def fetch_component_dependencies(self) -> "MintHelper":
        """Verify component dependencies."""
        if self.chain_type != ChainType.ETHEREUM:
            self.dependencies = []
            return self

        dependencies = []
        try:
            for ptype in PackageType:
                if not hasattr(self.package_configuration, ptype.to_plural()):
                    continue
                for dependency in getattr(
                    self.package_configuration, ptype.to_plural()
                ):
                    units = self.subgraph.get_record_by_package_id(
                        package_id=PackageId(package_type=ptype, public_id=dependency)
                    ).get("units", [])
                    if len(units) == 0:
                        raise click.ClickException(
                            f"Could not find on-chain token for {dependency} of type {ptype}"
                        )
                    unit, *_ = sorted(units, key=lambda x: x["tokenId"])
                    dependencies.append(unit["tokenId"])
            self.dependencies = sorted(list(map(int, dependencies)))
        except RequestConnectionError as e:
            raise click.ClickException(message=f"Error interacting with subgraph; {e}")
        return self

    def verify_service_dependencies(self, agent_id: int) -> "MintHelper":
        """Verify component dependencies."""
        self.agent_id = agent_id
        if self.chain_type == ChainType.LOCAL:
            return self

        try:
            units = self.subgraph.get_component_by_token(
                token_id=agent_id,
                package_type=PackageType.AGENT,
            ).get("units", [])
        except RequestConnectionError as e:
            raise click.ClickException(message=f"Error interacting with subgraph; {e}")
        if len(units) == 0:
            raise click.ClickException(f"No agents found with token ID {agent_id}")

        unit, *_ = sorted(units, key=lambda x: x["tokenId"])
        expected = PublicId.from_str(unit["publicId"]).to_any()
        agent = cast(Service, self.package_configuration).agent.to_any()
        if expected != agent:
            raise click.ClickException(
                f"Public ID `{expected}` for token {agent_id} does not match with the one "
                f"defained in the service package `{agent}`"
            )
        return self

    def publish_metadata(self) -> "MintHelper":
        """Publish metadata."""
        self.metadata_hash, self.metadata_string = publish_metadata(
            package_id=self.package_configuration.package_id,
            package_path=self.package_path,
            nft=cast(str, self.nft),
            description=self.package_configuration.description,
        )
        return self

    def mint_component(
        self,
        owner: Optional[str] = None,
        component_type: UnitType = UnitType.COMPONENT,
    ) -> None:
        """Mint component."""

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.registries_manager,
                (
                    ContractConfigs.component_registry
                    if component_type == UnitType.COMPONENT
                    else ContractConfigs.agent_registry
                ),
            )
        )

        try:
            self.token_id = self.manager.mint_component(
                metadata_hash=self.metadata_hash,
                component_type=component_type,
                owner=owner,
                dependencies=self.dependencies,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Component mint failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return

        click.echo("Component minted with:")
        click.echo(f"\tPublic ID: {self.package_configuration.public_id}")
        click.echo(f"\tMetadata Hash: {self.metadata_hash}")
        if self.token_id is not None:
            click.echo(f"\tToken ID: {self.token_id}")
            (Path.cwd() / f"{self.token_id}.json").write_text(self.metadata_string)
        else:  # pragma: nocover
            raise click.ClickException(
                "Could not verify metadata hash to retrieve the token ID"
            )

    def mint_agent(
        self,
        owner: Optional[str] = None,
    ) -> None:
        """Mint agent."""
        self.mint_component(
            owner=owner,
            component_type=UnitType.AGENT,
        )

    def mint_service(
        self,
        number_of_slots: int,
        cost_of_bond: int,
        threshold: Optional[int] = None,
        token: Optional[str] = None,
        owner: Optional[str] = None,
    ) -> None:
        """Mint service"""

        if (
            not is_service_manager_token_compatible_chain(ledger_api=self.ledger_api)
            and token is not None
        ):
            raise click.ClickException(
                "Cannot use custom token for bonding on L2 chains"
            )

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
            )
        )

        try:
            token_id = self.manager.mint_service(
                metadata_hash=self.metadata_hash,
                agent_ids=[
                    self.agent_id,
                ],
                number_of_slots_per_agent=[
                    number_of_slots,
                ],
                cost_of_bond_per_agent=[
                    cost_of_bond,
                ],
                threshold=threshold,
                token=token,
                owner=owner,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Component mint failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return

        click.echo("Service minted with:")
        click.echo(f"\tPublic ID: {self.package_configuration.public_id}")
        click.echo(f"\tMetadata Hash: {self.metadata_hash}")
        if token_id is not None:
            click.echo(f"\tToken ID: {token_id}")
            (Path.cwd() / f"{token_id}.json").write_text(self.metadata_string)
        else:  # pragma: nocover
            raise click.ClickException(
                "Could not verify metadata hash to retrieve the token ID"
            )

    def update_component(self, component_type: UnitType = UnitType.COMPONENT) -> None:
        """Update component."""
        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.registries_manager,
                (
                    ContractConfigs.component_registry
                    if component_type == UnitType.COMPONENT
                    else ContractConfigs.agent_registry
                ),
            )
        )
        try:
            self.token_id = self.manager.update_component(
                metadata_hash=self.metadata_hash,
                unit_id=cast(int, self.update_token),
                component_type=component_type,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Component update failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return

        click.echo("Component hash updated:")
        click.echo(f"\tPublic ID: {self.package_configuration.public_id}")
        click.echo(f"\tMetadata Hash: {self.metadata_hash}")
        if self.token_id is not None:
            click.echo(f"\tToken ID: {self.token_id}")
            (Path.cwd() / f"{self.token_id}.json").write_text(self.metadata_string)
        else:  # pragma: nocover
            raise click.ClickException(
                "Could not verify metadata hash to retrieve the token ID"
            )

    def update_agent(self) -> None:
        """Update agent."""
        self.update_component(component_type=UnitType.AGENT)

    def update_service(
        self,
        number_of_slots: int,
        cost_of_bond: int,
        threshold: Optional[int] = None,
        token: Optional[str] = None,
    ) -> None:
        """Update service"""

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
            )
        )

        *_, state, _ = get_service_info(
            ledger_api=self.ledger_api,
            chain_type=self.chain_type,
            token_id=cast(int, self.update_token),
        )

        if ServiceState(state) != ServiceState.PRE_REGISTRATION:
            raise click.ClickException(
                "Cannot update service hash, service needs to be in the pre-registration state"
            )

        try:
            token_id = self.manager.update_service(
                metadata_hash=self.metadata_hash,
                service_id=cast(int, self.update_token),
                agent_ids=[
                    self.agent_id,
                ],
                number_of_slots_per_agent=[
                    number_of_slots,
                ],
                cost_of_bond_per_agent=[
                    cost_of_bond,
                ],
                threshold=threshold,
                token=token,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Component mint failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return

        click.echo("Service updated with:")
        click.echo(f"\tPublic ID: {self.package_configuration.public_id}")
        click.echo(f"\tMetadata Hash: {self.metadata_hash}")
        if token_id is not None:
            click.echo(f"\tToken ID: {token_id}")
            (Path.cwd() / f"{token_id}.json").write_text(self.metadata_string)
        else:  # pragma: nocover
            raise click.ClickException(
                "Could not verify metadata hash to retrieve the token ID"
            )


class ServiceHelper(OnChainHelper):
    """Service helper."""

    token: Optional[str]
    token_secured: bool

    def __init__(  # pylint: disable=too-many-arguments
        self,
        service_id: int,
        chain_type: ChainType,
        key: Optional[Path] = None,
        password: Optional[str] = None,
        hwi: bool = False,
        timeout: Optional[float] = None,
        retries: Optional[int] = None,
        sleep: Optional[float] = None,
        dry_run: bool = False,
    ) -> None:
        """Initialize object."""
        self.service_id = service_id
        super().__init__(
            chain_type,
            key,
            password,
            hwi,
            timeout=timeout,
            retries=retries,
            sleep=sleep,
            dry_run=dry_run,
        )
        self.manager = ServiceManager(
            ledger_api=self.ledger_api,
            crypto=self.crypto,
            chain_type=self.chain_type,
            timeout=self.timeout,
            retries=self.retries,
            sleep=self.sleep,
            dry_run=dry_run,
        )

    def check_is_service_token_secured(
        self,
        token: Optional[str] = None,
    ) -> "ServiceHelper":
        """Check if service"""
        if not is_service_manager_token_compatible_chain(ledger_api=self.ledger_api):
            self.token = token
            self.token_secured = False
            return self

        if self.chain_type == ChainType.CUSTOM:  # pragma: nocover
            self.check_required_enviroment_variables(
                configs=(ContractConfigs.service_registry_token_utility,)
            )

        self.token = token
        self.token_secured = is_service_token_secured(
            ledger_api=self.ledger_api,
            chain_type=self.chain_type,
            service_id=self.service_id,
        )
        if self.token_secured and self.token is None:
            raise click.ClickException(
                "Service is token secured, please provice token address using `--token` flag"
            )
        ContractConfigs.erc20.contracts[self.chain_type] = cast(str, self.token)
        return self

    def approve_erc20_usage(self, amount: int, spender: str) -> "ServiceHelper":
        """Approve ERC20 usage."""

        try:
            approve_erc20_usage(
                ledger_api=self.ledger_api,
                crypto=self.crypto,
                chain_type=self.chain_type,
                spender=spender,
                amount=amount,
                sender=self.crypto.address,
                dry_run=self.dry_run,
                timeout=self.timeout,
                sleep=self.sleep,
                retries=self.retries,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(f"Error getting approval : {e}")
        return self

    def activate_service(self) -> None:
        """Activate on-chain service"""

        if self.token_secured:
            amount = get_token_deposit_amount(
                self.ledger_api,
                chain_type=self.chain_type,
                service_id=self.service_id,
            )
            spender = ContractConfigs.get(
                name=SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT.name
            ).contracts[self.chain_type]
            self.approve_erc20_usage(
                amount=amount,
                spender=spender,
            )

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
            )
        )

        try:
            self.manager.activate(service_id=self.service_id)
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Service activation failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return
        click.echo("Service activated succesfully")

    def register_instance(self, instances: List[str], agent_ids: List[int]) -> None:
        """Register agents instances on an activated service"""

        if self.token_secured:
            amount = get_activate_registration_amount(
                ledger_api=self.ledger_api,
                chain_type=self.chain_type,
                service_id=self.service_id,
                agents=agent_ids,
            )
            spender = ContractConfigs.get(
                name=SERVICE_REGISTRY_TOKEN_UTILITY_CONTRACT.name
            ).contracts[self.chain_type]
            self.approve_erc20_usage(
                amount=amount,
                spender=spender,
            )

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
            )
        )

        try:
            self.manager.register_instance(
                service_id=self.service_id,
                instances=instances,
                agent_ids=agent_ids,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Service activation failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return
        click.echo("Agent instance registered succesfully")

    def deploy_service(
        self,
        reuse_multisig: bool = False,
        fallback_handler: Optional[str] = None,
    ) -> None:
        """Deploy a service with registration activated"""

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
                ContractConfigs.gnosis_safe_proxy_factory,
                ContractConfigs.gnosis_safe_same_address_multisig,
                ContractConfigs.multisend,
            )
        )

        try:
            self.manager.deploy(
                service_id=self.service_id,
                reuse_multisig=reuse_multisig,
                fallback_handler=fallback_handler,
            )
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Service deployment failed with following error; {e.__class__.__name__}({e})"
            ) from e

        if self.dry_run:  # pragma: nocover
            return
        click.echo("Service deployed succesfully")

    def terminate_service(self) -> None:
        """Terminate a service"""

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
            )
        )

        try:
            self.manager.terminate(service_id=self.service_id)
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Service terminatation failed with following error; {e.__class__.__name__}({e})"
            ) from e
        click.echo("Service terminated succesfully")

    def unbond_service(self) -> None:
        """Unbond a service"""

        self.check_required_enviroment_variables(
            configs=(
                ContractConfigs.service_manager,
                ContractConfigs.service_registry,
            )
        )

        try:
            self.manager.unbond(service_id=self.service_id)
        except ChainInteractionError as e:  # pragma: nocover
            raise click.ClickException(
                f"Service unbonding failed with following error; {e.__class__.__name__}({e})"
            ) from e
        click.echo("Service unbonded succesfully")


def print_service_info(service_id: int, chain_type: ChainType) -> None:
    """Print service information"""
    ledger_api, _ = OnChainHelper.get_ledger_and_crypto_objects(chain_type=chain_type)
    (
        security_deposit,
        multisig_address,
        _,
        threshold,
        max_agents,
        number_of_agent_instances,
        _service_state,
        cannonical_agents,
    ) = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )
    service_state = ServiceState(_service_state)
    rows = [
        ("Property", "Value"),
        ("Service State", service_state.name),
        ("Security Deposit", security_deposit),
        ("Multisig Address", multisig_address),
        ("Cannonical Agents", ", ".join(map(str, cannonical_agents))),
        ("Max Agents", max_agents),
        ("Threshold", threshold),
        ("Number Of Agent Instances", number_of_agent_instances),
    ]

    if service_state.value >= ServiceState.ACTIVE_REGISTRATION.value:
        rows.append(
            (
                "Registered Instances",
                "\n".join(
                    map(
                        lambda x: f"- {x}",
                        get_agent_instances(
                            ledger_api=ledger_api,
                            chain_type=chain_type,
                            token_id=service_id,
                        ).get("agentInstances", []),
                    ),
                ),
            ),
        )
    click.echo(Texttable().add_rows(rows=rows).draw())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2024 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

# nosec

"""Query templates."""

FIND_BY_PACKAGE_HASH = """
query getUnit {{
  units(where:{{packageHash:"{package_hash}"}}){{
    tokenId
    packageHash
    publicId
  }}
}}
"""

FIND_BY_PUBLIC_ID = """
query getUnit {{
  units(where:{{publicId: "{public_id}",packageType:{package_type}}}){{
    tokenId
    packageHash
    publicId
  }}
}}
"""

# The bandit check fails here and adding nosec tag does not work because
# the black formatter puts the nosec tag to the end of the declaration
# so I declared the template as singleton and deconstructed the tuple to
# constant value

(FIND_BY_TOKEN_ID,) = (  # nosec
    """
query getUnit {{
  units(where:{{tokenId: "{token_id}",packageType:{package_type}}}){{
    tokenId
    packageHash
    publicId
  }}
}}
""",
)
